"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""


# TO-DO implement all the concepts discussed in
# https://stackoverflow.com/questions/30081275/why-is-1000000000000000-in-range1000000000000001-so-fast-in-python-3?rq=1

from qrisp import control, cx, perm_lock, perm_unlock, x
from qrisp.core import QuantumVariable
from qrisp.environments.quantum_conditionals import (
    ConditionEnvironment,
    quantum_condition,
)
from qrisp.logic_synthesis.truth_tables import TruthTable


class QuantumIterator:
    def __init__(self, quantum_condition, init_index=None):
        self.quantum_condition = quantum_condition

    def __iter__(self):
        self.quantum_condition.start_dumping()

        return self

    def __next__(self):
        raise Exception(
            "__next__ method for abstract quantum iterator class not defined"
        )


class qRange:
    def __init__(self, max_index_qf, create_index_qf=False):
        from qrisp.qtypes.quantum_float import QuantumFloat

        if not isinstance(max_index_qf, QuantumFloat):
            raise Exception("Can only create quantum iterators from quantum variables")

        if max_index_qf.exponent != 0:
            raise Exception(
                "qRange can only be intialized with integer quantum indices"
            )

        self.create_index_qf = create_index_qf
        self.max_index_qf = max_index_qf

    def __iter__(self):
        if self.create_index_qf:
            self.index_qf = self.max_index_qf.duplicate(
                qs=self.max_index_qf.qs, name="index_qf"
            )
            self.index_qf.init_from(self.max_index_qf)

        else:
            self.index_qf = self.max_index_qf

        if not self.index_qf.signed:
            self.index_qf.extend(1, position=-1)

        x(self.index_qf)
        self.index_qf += 1

        self.quantum_condition_env = control(self.index_qf[-1])

        self.c_index = -1

        perm_lock(self.index_qf)

        return self

    def __next__(self):
        perm_unlock(self.index_qf)
        self.c_index += 1

        if self.c_index != 0:
            self.quantum_condition_env.__exit__(None, None, None)

            if self.c_index >= 2 ** (self.index_qf.size - 1):
                if self.create_index_qf:
                    x(self.index_qf)
                    self.index_qf += self.c_index - 1

                    if not self.max_index_qf.signed:
                        cx(self.max_index_qf, self.index_qf[:-1])
                    else:
                        cx(self.max_index_qf, self.index_qf)
                    self.index_qf.delete()
                else:
                    x(self.index_qf)
                    self.index_qf += self.c_index - 1

                    self.index_qf.reduce(self.index_qf[-1])

                perm_unlock(self.index_qf)
                raise StopIteration

            self.index_qf += 1
        # @quantum_condition
        # def neg_condition(sign_qubit):
        #     return sign_qubit

        # self.quantum_condition_env = neg_condition(self.index_qf[-1])

        self.quantum_condition_env.__enter__()
        perm_lock(self.index_qf)
        return int(self.c_index)



"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""

#
# Autogenerated by Thrift Compiler (0.15.0)
#
# DO NOT EDIT UNLESS YOU ARE SURE THAT YOU KNOW WHAT YOU ARE DOING
#
#  options string: py
#

import logging
import sys

from thrift.protocol.TProtocol import TProtocolException
from thrift.Thrift import (
    TApplicationException,
    TException,
    TFrozenDict,
    TMessageType,
    TProcessor,
    TType,
)
from thrift.transport import TTransport
from thrift.TRecursive import fix_spec

from .ttypes import *

all_structs = []


class Iface(object):
    def run(self, qc, shots, token):
        """
        Parameters:
         - qc
         - shots
         - token

        """
        pass

    def ping(self):
        pass


class Client(Iface):
    def __init__(self, iprot, oprot=None):
        self._iprot = self._oprot = iprot
        if oprot is not None:
            self._oprot = oprot
        self._seqid = 0

    def run(self, qc, shots, token):
        """
        Parameters:
         - qc
         - shots
         - token

        """
        self.send_run(qc, shots, token)
        return self.recv_run()

    def send_run(self, qc, shots, token):
        self._oprot.writeMessageBegin("run", TMessageType.CALL, self._seqid)
        args = run_args()
        args.qc = qc
        args.shots = shots
        args.token = token
        args.write(self._oprot)
        self._oprot.writeMessageEnd()
        self._oprot.trans.flush()

    def recv_run(self):
        iprot = self._iprot
        (fname, mtype, rseqid) = iprot.readMessageBegin()
        if mtype == TMessageType.EXCEPTION:
            x = TApplicationException()
            x.read(iprot)
            iprot.readMessageEnd()
            raise x
        result = run_result()
        result.read(iprot)
        iprot.readMessageEnd()
        if result.success is not None:
            return result.success
        raise TApplicationException(
            TApplicationException.MISSING_RESULT, "run failed: unknown result"
        )

    def ping(self):
        self.send_ping()
        return self.recv_ping()

    def send_ping(self):
        self._oprot.writeMessageBegin("ping", TMessageType.CALL, self._seqid)
        args = ping_args()
        args.write(self._oprot)
        self._oprot.writeMessageEnd()
        self._oprot.trans.flush()

    def recv_ping(self):
        iprot = self._iprot
        (fname, mtype, rseqid) = iprot.readMessageBegin()
        if mtype == TMessageType.EXCEPTION:
            x = TApplicationException()
            x.read(iprot)
            iprot.readMessageEnd()
            raise x
        result = ping_result()
        result.read(iprot)
        iprot.readMessageEnd()
        if result.success is not None:
            return result.success
        raise TApplicationException(
            TApplicationException.MISSING_RESULT, "ping failed: unknown result"
        )


class Processor(Iface, TProcessor):
    def __init__(self, handler):
        self._handler = handler
        self._processMap = {}
        self._processMap["run"] = Processor.process_run
        self._processMap["ping"] = Processor.process_ping
        self._on_message_begin = None

    def on_message_begin(self, func):
        self._on_message_begin = func

    def process(self, iprot, oprot):
        (name, type, seqid) = iprot.readMessageBegin()
        if self._on_message_begin:
            self._on_message_begin(name, type, seqid)
        if name not in self._processMap:
            iprot.skip(TType.STRUCT)
            iprot.readMessageEnd()
            x = TApplicationException(
                TApplicationException.UNKNOWN_METHOD, "Unknown function %s" % (name)
            )
            oprot.writeMessageBegin(name, TMessageType.EXCEPTION, seqid)
            x.write(oprot)
            oprot.writeMessageEnd()
            oprot.trans.flush()
            return
        else:
            self._processMap[name](self, seqid, iprot, oprot)
        return True

    def process_run(self, seqid, iprot, oprot):
        args = run_args()
        args.read(iprot)
        iprot.readMessageEnd()
        result = run_result()
        try:
            result.success = self._handler.run(args.qc, args.shots, args.token)
            msg_type = TMessageType.REPLY
        except TTransport.TTransportException:
            raise
        except TApplicationException as ex:
            logging.exception("TApplication exception in handler")
            msg_type = TMessageType.EXCEPTION
            result = ex
        except Exception:
            logging.exception("Unexpected exception in handler")
            msg_type = TMessageType.EXCEPTION
            result = TApplicationException(
                TApplicationException.INTERNAL_ERROR, "Internal error"
            )
        oprot.writeMessageBegin("run", msg_type, seqid)
        result.write(oprot)
        oprot.writeMessageEnd()
        oprot.trans.flush()

    def process_ping(self, seqid, iprot, oprot):
        args = ping_args()
        args.read(iprot)
        iprot.readMessageEnd()
        result = ping_result()
        try:
            result.success = self._handler.ping()
            msg_type = TMessageType.REPLY
        except TTransport.TTransportException:
            raise
        except TApplicationException as ex:
            logging.exception("TApplication exception in handler")
            msg_type = TMessageType.EXCEPTION
            result = ex
        except Exception:
            logging.exception("Unexpected exception in handler")
            msg_type = TMessageType.EXCEPTION
            result = TApplicationException(
                TApplicationException.INTERNAL_ERROR, "Internal error"
            )
        oprot.writeMessageBegin("ping", msg_type, seqid)
        result.write(oprot)
        oprot.writeMessageEnd()
        oprot.trans.flush()


# HELPER FUNCTIONS AND STRUCTURES


class run_args(object):
    """
    Attributes:
     - qc
     - shots
     - token

    """

    def __init__(
        self,
        qc=None,
        shots=None,
        token=None,
    ):
        self.qc = qc
        self.shots = shots
        self.token = token

    def read(self, iprot):
        if (
            iprot._fast_decode is not None
            and isinstance(iprot.trans, TTransport.CReadableTransport)
            and self.thrift_spec is not None
        ):
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.STRUCT:
                    self.qc = QuantumCircuit()
                    self.qc.read(iprot)
                else:
                    iprot.skip(ftype)
            elif fid == 2:
                if ftype == TType.I32:
                    self.shots = iprot.readI32()
                else:
                    iprot.skip(ftype)
            elif fid == 3:
                if ftype == TType.STRING:
                    self.token = (
                        iprot.readString().decode("utf-8", errors="replace")
                        if sys.version_info[0] == 2
                        else iprot.readString()
                    )
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(
                oprot._fast_encode(self, [self.__class__, self.thrift_spec])
            )
            return
        oprot.writeStructBegin("run_args")
        if self.qc is not None:
            oprot.writeFieldBegin("qc", TType.STRUCT, 1)
            self.qc.write(oprot)
            oprot.writeFieldEnd()
        if self.shots is not None:
            oprot.writeFieldBegin("shots", TType.I32, 2)
            oprot.writeI32(self.shots)
            oprot.writeFieldEnd()
        if self.token is not None:
            oprot.writeFieldBegin("token", TType.STRING, 3)
            oprot.writeString(
                self.token.encode("utf-8") if sys.version_info[0] == 2 else self.token
            )
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ["%s=%r" % (key, value) for key, value in self.__dict__.items()]
        return "%s(%s)" % (self.__class__.__name__, ", ".join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)


all_structs.append(run_args)
run_args.thrift_spec = (
    None,  # 0
    (
        1,
        TType.STRUCT,
        "qc",
        [QuantumCircuit, None],
        None,
    ),  # 1
    (
        2,
        TType.I32,
        "shots",
        None,
        None,
    ),  # 2
    (
        3,
        TType.STRING,
        "token",
        "UTF8",
        None,
    ),  # 3
)


class run_result(object):
    """
    Attributes:
     - success

    """

    def __init__(
        self,
        success=None,
    ):
        self.success = success

    def read(self, iprot):
        if (
            iprot._fast_decode is not None
            and isinstance(iprot.trans, TTransport.CReadableTransport)
            and self.thrift_spec is not None
        ):
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 0:
                if ftype == TType.MAP:
                    self.success = {}
                    (_ktype64, _vtype65, _size63) = iprot.readMapBegin()
                    for _i67 in range(_size63):
                        _key68 = (
                            iprot.readString().decode("utf-8", errors="replace")
                            if sys.version_info[0] == 2
                            else iprot.readString()
                        )
                        _val69 = iprot.readI32()
                        self.success[_key68] = _val69
                    iprot.readMapEnd()
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(
                oprot._fast_encode(self, [self.__class__, self.thrift_spec])
            )
            return
        oprot.writeStructBegin("run_result")
        if self.success is not None:
            oprot.writeFieldBegin("success", TType.MAP, 0)
            oprot.writeMapBegin(TType.STRING, TType.I32, len(self.success))
            for kiter70, viter71 in self.success.items():
                oprot.writeString(
                    kiter70.encode("utf-8") if sys.version_info[0] == 2 else kiter70
                )
                oprot.writeI32(viter71)
            oprot.writeMapEnd()
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ["%s=%r" % (key, value) for key, value in self.__dict__.items()]
        return "%s(%s)" % (self.__class__.__name__, ", ".join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)


all_structs.append(run_result)
run_result.thrift_spec = (
    (
        0,
        TType.MAP,
        "success",
        (TType.STRING, "UTF8", TType.I32, None, False),
        None,
    ),  # 0
)


class ping_args(object):
    def read(self, iprot):
        if (
            iprot._fast_decode is not None
            and isinstance(iprot.trans, TTransport.CReadableTransport)
            and self.thrift_spec is not None
        ):
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(
                oprot._fast_encode(self, [self.__class__, self.thrift_spec])
            )
            return
        oprot.writeStructBegin("ping_args")
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ["%s=%r" % (key, value) for key, value in self.__dict__.items()]
        return "%s(%s)" % (self.__class__.__name__, ", ".join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)


all_structs.append(ping_args)
ping_args.thrift_spec = ()


class ping_result(object):
    """
    Attributes:
     - success

    """

    def __init__(
        self,
        success=None,
    ):
        self.success = success

    def read(self, iprot):
        if (
            iprot._fast_decode is not None
            and isinstance(iprot.trans, TTransport.CReadableTransport)
            and self.thrift_spec is not None
        ):
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 0:
                if ftype == TType.STRUCT:
                    self.success = BackendStatus()
                    self.success.read(iprot)
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(
                oprot._fast_encode(self, [self.__class__, self.thrift_spec])
            )
            return
        oprot.writeStructBegin("ping_result")
        if self.success is not None:
            oprot.writeFieldBegin("success", TType.STRUCT, 0)
            self.success.write(oprot)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ["%s=%r" % (key, value) for key, value in self.__dict__.items()]
        return "%s(%s)" % (self.__class__.__name__, ", ".join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)


all_structs.append(ping_result)
ping_result.thrift_spec = (
    (
        0,
        TType.STRUCT,
        "success",
        [BackendStatus, None],
        None,
    ),  # 0
)
fix_spec(all_structs)
del all_structs



"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""


import numpy as np
import sympy

import qrisp.circuit.standard_operations as ops
from qrisp.circuit import Clbit, Instruction, Operation, Qubit

# Class to describe quantum circuits
# The naming of the attributes is rather similar to the qiskit equivalent
# in order to allow compatibility of qiskit programs to qrisp
# The key attributes are

# The list of qubits (.qubits).
# the list of classical bits (.clbits)
# the list of instructions (.data)


class QuantumCircuit:
    """
    This class describes quantum circuits. Many of the attribute and method names are
    oriented at the `Qiskit QuantumCircuit
    <https://qiskit.org/documentation/stubs/qiskit.circuit.QuantumCircuit.html>`_ class
    in order to provide a high degree of compatibility.

    QuantumCircuits can be visualized by calling ``print`` on them.

    Qrisp QuantumCircuits can be quickly generated out of existing Qiskit
    QuantumCircuits with the :meth:`from_qiskit <qrisp.QuantumCircuit.from_qiskit>`
    method.


    Parameters
    ----------

    num_qubits : integer, optional
        The amount of qubits, this QuantumCircuit is initialized with. The default is 0.
    num_clbits : integer, optional
        The amount of classical bits. The default is 0.
    name : string, optional
        A name for the QuantumCircuit. The default will generated a generic name.


    Examples
    --------

    We create a QuantumCircuit containing a so-called fan-out gate:

    >>> from qrisp import QuantumCircuit
    >>> qc_0 = QuantumCircuit(4, name = "fan out")
    >>> qc_0.cx(0, range(1,4))
    >>> print(qc_0)
    qb_0: ──■────■────■──
          ┌─┴─┐  │    │
    qb_1: ┤ X ├──┼────┼──
          └───┘┌─┴─┐  │
    qb_2: ─────┤ X ├──┼──
               └───┘┌─┴─┐
    qb_3: ──────────┤ X ├
                    └───┘

    Note that the :meth:`cx gate appending method <qrisp.QuantumCircuit.cx>` (like all
    other gate appending methods) can be called with integers, Qubit objects,
    lists of integers or lists of Qubit objects.

    We now turn this QuantumCircuit into a gate and append to another QuantumCircuit to
    generate a GHZ state:

    >>> qc_1 = QuantumCircuit(4)
    >>> qc_1.h(0)
    >>> qc_1.append(qc_0.to_gate(), qc_1.qubits)
    >>> print(qc_1)
          ┌───┐┌──────────┐
    qb_4: ┤ H ├┤0         ├
          └───┘│          │
    qb_5: ─────┤1         ├
               │  fan out │
    qb_6: ─────┤2         ├
               │          │
    qb_7: ─────┤3         ├
               └──────────┘

    Finally, we add a measurement and evaluate the circuit:

    >>> qc_1.measure(qc_1.qubits)
    >>> print(qc_1)
          ┌───┐┌──────────┐┌─┐
    qb_4: ┤ H ├┤0         ├┤M├─────────
          └───┘│          │└╥┘┌─┐
    qb_5: ─────┤1         ├─╫─┤M├──────
               │  fan out │ ║ └╥┘┌─┐
    qb_6: ─────┤2         ├─╫──╫─┤M├───
               │          │ ║  ║ └╥┘┌─┐
    qb_7: ─────┤3         ├─╫──╫──╫─┤M├
               └──────────┘ ║  ║  ║ └╥┘
    cb_0: ══════════════════╩══╬══╬══╬═
                               ║  ║  ║
    cb_1: ═════════════════════╩══╬══╬═
                                  ║  ║
    cb_2: ════════════════════════╩══╬═
                                     ║
    cb_3: ═══════════════════════════╩═
    >>> qc_1.run(shots = 1000)
    {'0000': 500, '1111': 500}

    **Converting from Qiskit**

    We construct the very same fan out QuantumCircuit in Qiskit:

    >>> from qiskit import QuantumCircuit as QiskitQuantumCircuit
    >>> qc_2 = QiskitQuantumCircuit(4)
    >>> qc_2.cx(0, range(1,4))
    >>> print(qc_2)
    q_0: ──■────■────■──
         ┌─┴─┐  │    │
    q_1: ┤ X ├──┼────┼──
         └───┘┌─┴─┐  │
    q_2: ─────┤ X ├──┼──
              └───┘┌─┴─┐
    q_3: ──────────┤ X ├
                   └───┘

    To acquire the Qrisp QuantumCircuit we call the
    :meth:`from_qiskit <qrisp.QuantumCircuit.from_qiskit>` method. Note that we don't
    need to create a QuantumCircuit object first as this is a class method.

    >>> qrisp_qc_2 = QuantumCircuit.from_qiskit(qc_2)
    >>> print(qrisp_qc_2)
     qb_8: ──■────■────■──
           ┌─┴─┐  │    │
     qb_9: ┤ X ├──┼────┼──
           └───┘┌─┴─┐  │
    qb_10: ─────┤ X ├──┼──
                └───┘┌─┴─┐
    qb_11: ──────────┤ X ├
                     └───┘

    **Abstract Parameters**

    Abstract parameters are represented by `Sympy symbols
    <https://docs.sympy.org/latest/modules/core.html#module-sympy.core.symbol>`_
    in Qrisp.

    We create a QuantumCircuit with some abstract parameters and bind them subsequently.

    >>> from qrisp import QuantumCircuit
    >>> from sympy import symbols
    >>> qc = QuantumCircuit(3)

    Create some Sympy symbols and use them as abstract parameters for phase gates:

    >>> abstract_parameters = symbols("a b c")
    >>> for i in range(3): qc.p(abstract_parameters[i], i)

    Create the substitution dictionary and bind the parameters:

    >>> subs_dic = {abstract_parameters[i] : i for i in range(3)}
    >>> bound_qc = qc.bind_parameters(subs_dic)
    >>> print(bound_qc)
          ┌──────┐
    qb_0: ┤ P(0) ├
          ├──────┤
    qb_1: ┤ P(1) ├
          ├──────┤
    qb_2: ┤ P(2) ├
          └──────┘

    """

    qubit_index_counter = np.zeros(1, dtype=int)
    fast_append = False

    def __init__(self, num_qubits=0, num_clbits=0, name=None):
        object.__setattr__(self, "data", [])
        object.__setattr__(self, "qubits", [])
        object.__setattr__(self, "clbits", [])

        self.abstract_params = set([])

        if name is None:
            self.name = "circuit_" + str(id(self))[-5:]
        else:
            self.name = name

        if isinstance(num_qubits, int):
            for i in range(num_qubits):
                self.qubit_index_counter[0] += 1
                self.qubits.append(Qubit("qb_" + str(self.qubit_index_counter[0])))
        else:
            raise Exception(
                f"Tried to initialize QuantumCircuit with type {type(num_qubits)}"
            )

        if isinstance(num_clbits, int):
            for i in range(num_clbits):
                self.add_clbit()
        else:
            raise Exception(
                f"Tried to initialize QuantumCircuit with type {type(num_clbits)}"
            )

        self.last_qubit_count = -1

    # Method to add qubit objects to the circuit
    def add_qubit(self, qubit=None):
        """
        Adds a Qubit to the QuantumCircuit.

        Parameters
        ----------
        qubit : Qubit, optional
            The Qubit to be added. If given none, a new Qubit will be generated.

        Returns
        -------
        Qubit
            The added Qubit.

        Examples
        --------

        >>> from qrisp import QuantumCircuit
        >>> qc = QuantumCircuit()
        >>> qc.add_qubit()
        >>> qc.qubits
        [qb_0]

        """

        self.qubit_index_counter += 1
        if qubit is None:
            qubit = Qubit("qb_" + str(self.qubit_index_counter[0]))

        if not isinstance(qubit, Qubit):
            raise Exception(f"Tried to add type {type(qubit)} as a qubit")

        for qb in self.qubits:
            if qb.identifier == qubit.identifier:
                raise Exception(f"Qubit name {qubit.identifier} already exists")

        self.qubits.append(qubit)

        return self.qubits[-1]

    # Method to add classical bit objects to the circuit
    def add_clbit(self, clbit=None):
        """
        Adds a classical bit to the QuantumCircuit.

        Parameters
        ----------
        clbit : Clbit, optional
            The classical bit to be added. If given none, a new Clbit will be generated.

        Returns
        -------
        Clbit
            The added Clbit.

        """

        if clbit is None:
            clbit = Clbit("cb_" + str(len(self.clbits)))

        if not isinstance(clbit, Clbit):
            raise Exception(f"Tried to add type {type(clbit)} as a classical bit")

        for cb in self.clbits:
            if cb.identifier == clbit.identifier:
                raise Exception(f"Clbit name {clbit.identifier} already exists")

        self.clbits.append(clbit)

        return self.clbits[-1]

    # Method to transform the given circuit into an operation object
    def to_op(self, name=None):
        """
        Method to return an Operation object generated out of this QuantumCircuit.

        Operation objects can be appended to other QuantumCircuits.

        An alias for Qiskit compatibility is the
        :meth:`to_gate<qrisp.QuantumCircuit.to_gate>` method.

        Parameters
        ----------
        name : string, optional
            The name of the gate. By default, the QuantumCircuit's name will be used.

        Returns
        -------
        Operation
            The Operation defined by this QuantumCircuit.

        Examples
        --------

        >>> from qrisp import QuantumCircuit
        >>> qc_0 = QuantumCircuit(4)
        >>> qc_0.x(qc.qubits)
        >>> operation = qc_0.to_gate()
        >>> qc_1 = QuantumCircuit(4)
        >>> qc_1.append(operation, qc_1.qubits)

        """

        if name is None:
            name = self.name

        definition = self.copy()
        i = 0

        while i < len(definition.data):
            if definition.data[i].op.name in ["qb_alloc", "qb_dealloc"]:
                definition.data.pop(i)
                continue
            i += 1

        return Operation(
            name=name,
            num_qubits=len(self.qubits),
            num_clbits=len(self.clbits),
            definition=definition,
            params=[],
        )

    # Wrapper to increase Qiskit compatibility
    def to_gate(self, name=None):
        """
        Similar to :meth:`to_op <qrisp.QuantumCircuit.to_op>` but raises an exception
        if self contains classical bits (like the
        `Qiskit equivalent
        <https://qiskit.org/documentation/stubs/qiskit.circuit.QuantumCircuit.to_gate.html>`_).
        # noqa

        Parameters
        ----------
        name : str, optional
            A name for the resulting gate. The default is None.

        Raises
        ------
        Exception
            Tried to turn a circuit including classical bits into unitary gate

        Returns
        -------
        Operation
            The QuantumCircuit turned into an :ref:`Operation` instance.

        """

        if len(self.clbits) != 0:
            raise Exception(
                "Tried to turn a circuit including classical bits into unitary gate"
            )
        return self.to_op(name)

    # Method to extend the given circuit with another circuit
    # The dic translation dic encodes how the qubits should be plugged into each other
    def extend(self, other, translation_dic="id"):
        """
        Extends self in-place by another QuantumCircuit.

        Parameters
        ----------
        other : QuantumCircuit
            The QuantumCircuit to extend by.
        translation_dic : dict, optional
            The dictionary containing the information about which Qubits and Clbits
            should be plugged into each other. This dictionary should contain qubits of
            other as keys and qubits of self as values. If given none, it is assumed
            that both QuantumCircuits have matching Qubits.


        Examples
        --------

        We create two QuantumCircuits and extend the first with reversed qubit order by
        the other:

        >>> from qrisp import QuantumCircuit
        >>> extension_qc = QuantumCircuit(4)
        >>> qc_to_extend = QuantumCircuit(4)
        >>> extension_qc.cx(0, 1)
        >>> extension_qc.cy(0, 2)
        >>> extension_qc.cz(0, 3)
        >>> print(extension_qc)
        qb_0: ──■────■────■──
              ┌─┴─┐  │    │
        qb_1: ┤ X ├──┼────┼──
              └───┘┌─┴─┐  │
        qb_2: ─────┤ Y ├──┼──
                   └───┘┌─┴─┐
        qb_3: ──────────┤ Z ├
                        └───┘
        >>> translation_dic = {extension_qc.qubits[i] : qc_to_extend.qubits[-1-i]
        >>> for i in range(4)}
        >>> qc_to_extend.extend(extension_qc, translation_dic)
        >>> print(qc_to_extend)
                        ┌───┐
        qb_4: ──────────┤ Z ├
                   ┌───┐└─┬─┘
        qb_5: ─────┤ Y ├──┼──
              ┌───┐└─┬─┘  │
        qb_6: ┤ X ├──┼────┼──
              └─┬─┘  │    │
        qb_7: ──■────■────■──

        """

        if translation_dic == "id":
            translation_dic = {}
            for qb in other.qubits:
                translation_dic[qb] = qb

            for cb in other.clbits:
                translation_dic[cb] = cb

        # Copy in order to prevent modification
        translation_dic = dict(translation_dic)

        for key in list(translation_dic.keys()):
            if isinstance(key, (Qubit, Clbit)):
                translation_dic[key.identifier] = translation_dic[key]

        for i in range(len(other.data)):
            instruction_other = other.data[i]
            qubits = []
            for qb in instruction_other.qubits:
                qubits.append(translation_dic[qb.identifier])

            clbits = []

            for cb in instruction_other.clbits:
                clbits.append(translation_dic[cb.identifier])

            self.append(instruction_other.op, qubits, clbits)

    # Returns a copy of self
    def copy(self):
        """
        Returns a copy of the given QuantumCircuit.

        Returns
        -------
        QuantumCircuit
            The copied QuantumCircuit.

        """
        # If an inital circuit is given we construct a new instance

        res = QuantumCircuit()

        object.__setattr__(res, "data", list(self.data))
        object.__setattr__(res, "qubits", list(self.qubits))
        object.__setattr__(res, "clbits", list(self.clbits))

        try:
            res.abstract_params = set(self.abstract_params)
        except AttributeError:
            pass

        return res

        return QuantumCircuit(init_qc=self)

    # Returns a copy of self but with no instructions
    def clearcopy(self):
        """
        Returns a copy of the given QuantumCircuit but without any data
        (i.e. just the Qubits and Clbits).

        Returns
        -------
        QuantumCircuit
            The empty, copied QuantumCircuit.

        """
        temp_data = list(self.data)
        self.data = []
        res = self.copy()
        self.data = temp_data
        return res

    # TO-DO write qiskit independent printer

    # Printing method
    def __str__(self):

        from qiskit.visualization.circuit_visualization import circuit_drawer

        from qrisp.interface.circuit_converter import convert_circuit

        try:
            res_str = str(
                circuit_drawer(
                    convert_circuit(self, target_api="qiskit", transpile=False),
                    style="text",
                    cregbundle=False,
                )
            )
        except AttributeError:
            raise Exception(
                "Tried to print QuantumSession with uncompiled QuantumEnvironments"
            )

        return res_str

    # Method which compares the unitary of two given circuits and returns
    # True if they are equivalent
    def compare_unitary(self, other, precision=4, ignore_gphase=False):
        """
        Compares the unitaries of two QuantumCircuits. This can be used to check if a
        QuantumCircuit transformation is valid.

        Parameters
        ----------
        other : QuantumCircuit
            The QuantumCircuit to compare to.
        precision : int, optional
            The precision of the comparison. This function will return True, if the norm
            of the difference of the unitaries is below the precision. The default is 4.
        ignore_gphase: bool, optional
            If set to True, this method returns True if the unitaries only differ in a
            global phase.

        Returns
        -------
        Bool
            The comparison outcome.

        Examples
        --------

        We create two QuantumCircuit with equivalent unitaries but differing by a
        non-trivial commutation:

        >>> from qrisp import QuantumCircuit
        >>> qc_0 = QuantumCircuit(2)
        >>> qc_1 = QuantumCircuit(2)
        >>> qc_0.z(0)
        >>> qc_0.cx(0,1)
        >>> print(qc_0)
              ┌───┐
        qb_0: ┤ Z ├──■──
              └───┘┌─┴─┐
        qb_1: ─────┤ X ├
                   └───┘
        >>> qc_1.cx(0,1)
        >>> qc_1.z(0)
        >>> print(qc_1)
                   ┌───┐
        qb_2: ──■──┤ Z ├
              ┌─┴─┐└───┘
        qb_3: ┤ X ├─────
              └───┘
        >>> qc_0.compare_unitary(qc_1)
        True


        """

        if len(self.qubits) != len(other.qubits):
            return False

        unitary_self = self.get_unitary()
        unitary_other = other.get_unitary()

        if ignore_gphase:
            arg_max = np.argmax(np.abs(unitary_self.flatten()))

            unitary_self = (
                unitary_self
                * unitary_other.flatten()[arg_max]
                / unitary_self.flatten()[arg_max]
            )

        from numpy.linalg import norm

        return bool(norm(unitary_self - unitary_other) < 10**-precision)

    # Converts several types of inputs to qubit lists.
    # Possible inputs are
    #
    # A qubit object
    # An integer
    # A list of integers
    # A list of qubits
    def convert_to_qubit_list(self, input, inner_recursion=False):
        if isinstance(input, Qubit):
            if inner_recursion:
                return input
            else:
                return [input]

        if isinstance(input, int):
            try:
                return self.convert_to_qubit_list(
                    self.qubits[input], inner_recursion=inner_recursion
                )
            except IndexError:
                raise Exception(
                    "Not enough qubits in circuit to access qubit " + str(input) + "."
                )

        if isinstance(input, list):
            return_list = []
            for qb in input:
                return_list.append(self.convert_to_qubit_list(qb, inner_recursion=True))
            return return_list

        raise Exception(
            "Could not convert input type " + type(input) + " to qubit list"
        )

    # Similar function as above but with classical bits
    def convert_to_clbit_list(self, input, inner_recursion=False):
        if isinstance(input, Clbit):
            if inner_recursion:
                return input
            else:
                return [input]

        if isinstance(input, int):
            try:
                return self.convert_to_clbit_list(
                    self.clbits[input], inner_recursion=inner_recursion
                )
            except IndexError:
                raise Exception(
                    "Not enough clbits in circuit to access clbit " + str(input) + "."
                )

        if isinstance(input, list):
            return_list = []
            for cb in input:
                return_list.append(self.convert_to_clbit_list(cb, inner_recursion=True))
            return return_list

        return self.convert_to_clbit_list(list(input))

    # Generates the inverse of self by applying the inverse gates in reversed order
    def inverse(self):
        """
        Returns the inverse/daggered QuantumCircuit.

        Returns
        -------
        inverted_circuit : QuantumCircuit
            The inverted QuantumCircuit.

        Examples
        --------

        Daggering a QuantumCircuit reverses the order and daggers each operation.

        >>> from qrisp import QuantumCircuit
        >>> import numpy as np
        >>> qc = QuantumCircuit(1)
        >>> qc.x(0)
        >>> qc.p(np.pi/2, 0)
        >>> qc.y(0)
        >>> print(qc.inverse())
                ┌───┐┌─────────┐┌───┐
          qb_0: ┤ Y ├┤ P(-π/2) ├┤ X ├
                └───┘└─────────┘└───┘

        For the phase gate, a daggering implies the reversal of the phase -
        Pauli gates however are invariant under daggering.

        """

        inverted_circuit = self.clearcopy()
        for instr in self.data[::-1]:
            inverted_circuit.append(instr.op.inverse(), instr.qubits, instr.clbits)

        return inverted_circuit

    # Generate the circuits unitary matrix
    def get_unitary(self, decimals=-1):
        """
        Acquires the unitary matrix of the given QuantumCircuit as a Numpy array.

        This method also works with abstract parameters. In this case a Numpy array
        with Sympy entries is returned.

        Parameters
        ----------
        decimals : integer, optional
            The amount of decimals to be rounded to. By default, the full precision is
            returned.

        Returns
        -------
        numpy.ndarray
            The unitary matrix as a numpy array.

        Examples
        --------

        We synthesize a controlled phase gate and inspect the unitary:

        >>> from qrisp import QuantumCircuit
        >>> import numpy as np
        >>> qc = QuantumCircuit(2)
        >>> phi = np.pi
        >>> qc.p(phi/2, 0)
        >>> qc.p(phi/2, 1)
        >>> qc.cx(0,1)
        >>> qc.p(-phi/2, 1)
        >>> qc.cx(0,1)
        >>> qc.get_unitary(decimals = 4)
        array([[ 1.+0.j,  0.+0.j,  0.+0.j,  0.+0.j],
               [ 0.+0.j,  1.+0.j,  0.+0.j,  0.+0.j],
               [ 0.+0.j,  0.+0.j,  1.+0.j,  0.+0.j],
               [ 0.+0.j,  0.+0.j,  0.+0.j, -1.+0.j]], dtype=complex64)

        We now synthesize the exact same QuantumCircuit but this time ``phi`` is a Sympy
        symbol.

        >>> from sympy import Symbol
        >>> qc = QuantumCircuit(2)
        >>> phi = Symbol("phi")
        >>> qc.p(phi/2, 0)
        >>> qc.p(phi/2, 1)
        >>> qc.cx(0,1)
        >>> qc.p(-phi/2, 1)
        >>> qc.cx(0,1)
        >>> qc.get_unitary(decimals = 4)
        array([[1, 0, 0, 0],
               [0, 1, 0, 0],
               [0, 0, 1, 0],
               [0, 0, 0, exp(I*phi)]], dtype=object)

        """
        from qrisp.simulator import calc_circuit_unitary

        res = calc_circuit_unitary(self)

        if decimals != -1:
            if res.dtype == np.dtype("O"):
                raveled_res = res.ravel()
                for i in range(len(raveled_res)):
                    expression = raveled_res[i]
                    for a in sympy.preorder_traversal(expression):
                        if isinstance(a, sympy.Float):
                            rounded_float = round(a, decimals)
                            if abs(rounded_float - 1) < 10**-decimals:
                                expression = expression.subs(a, 1)
                            else:
                                expression = expression.subs(a, rounded_float)

                    raveled_res[i] = expression
            else:
                res = np.round(res, decimals)

        return res

    def get_depth_dic(self):
        from qrisp.misc import get_depth_dic

        return get_depth_dic(self)

    def cnot_count(self):
        """
        Method to determine the amount of CNOT gates used in this QuantumCircuit.

        Returns
        -------
        int
            The amount of CNOT gates.

        """

        from qrisp.misc import cnot_count

        return cnot_count(self)

    def transpile(self, transpilation_level=np.inf, **qiskit_kwargs):
        """
        Transpiles the QuantumCircuit in the sense that there are no longer any
        synthesized gate objects. Furthermore, we can call the `Qiskit transpiler
        <https://qiskit.org/documentation/stubs/qiskit.compiler.transpile.html>`__
        by supplying keyword arguments.

        The Qiskit transpiler is not called, if no keyword arguments are given.

        Parameters
        ----------
        **qiskit_kwargs :
            Keyword arguments for the Qiskit transpiler.

        Returns
        -------
        QuantumCircuit
            The transpiled QuantumCircuit.

        Examples
        --------

        We create a QuantumCircuit and append a synthesized gate. Afterwards we
        transpile to a given set of basis gates using the Qiskit transpiler:

        >>> from qrisp import QuantumCircuit
        >>> qc = QuantumCircuit(3)
        >>> qc.mcx([0,1], 2)
        >>> print(qc)
        qb_0: ──■──
                │
        qb_1: ──■──
              ┌─┴─┐
        qb_2: ┤ X ├
              └───┘
        >>> print(qc.transpile(basis_gates = ["cx", "rz", "sx"]))
        global phase: 9π/8
              ┌──────────┐                 ┌───┐┌─────────┐   ┌───┐   ┌──────────┐┌───┐»
        qb_1: ┤ Rz(-π/4) ├─────────────────┤ X ├┤ Rz(π/4) ├───┤ X ├───┤ Rz(-π/4) ├┤ X ├»
              ├──────────┤                 └─┬─┘└─────────┘   └─┬─┘   └──────────┘└─┬─┘»
        qb_2: ┤ Rz(-π/4) ├───────────────────┼───────■──────────■──────────■────────┼──»
              ├─────────┬┘┌────┐┌─────────┐  │     ┌─┴─┐   ┌─────────┐   ┌─┴─┐      │  »
        qb_3: ┤ Rz(π/2) ├─┤ √X ├┤ Rz(π/2) ├──■─────┤ X ├───┤ Rz(π/4) ├───┤ X ├──────■──»
              └─────────┘ └────┘└─────────┘        └───┘   └─────────┘   └───┘         »
        «      ┌─────────┐┌───┐
        «qb_1: ┤ Rz(π/4) ├┤ X ├────────────
        «      └─────────┘└─┬─┘
        «qb_2: ─────────────■──────────────
        «      ┌─────────┐┌────┐┌─────────┐
        «qb_3: ┤ Rz(π/4) ├┤ √X ├┤ Rz(π/2) ├
        «      └─────────┘└────┘└─────────┘

        """
        from qrisp.circuit import transpile

        return transpile(self, transpilation_level, **qiskit_kwargs)

    # Counts the amount of operations self contains and returns
    # a dict {"operatio_name" : operation_count, ...}
    def count_ops(self):
        """
        Counts the amount of operations of each kind. Note that operations are
        identified by their name.

        Returns
        -------
        count_dic : dict
            A dictionary containing the gate counts.

        Examples
        --------

        We create a QuantumCircuit containing a number of gates and evaluates the
        gate-counts:

        >>> from qrisp import QuantumCircuit
        >>> qc = QuantumCircuit(5)
        >>> qc.x(qc.qubits)
        >>> qc.cx(0, range(1,5))
        >>> qc.z(1)
        >>> qc.t(range(5))
        >>> qc.count_ops()
        {'x': 5, 'cx': 4, 'z': 1, 't': 5}

        """
        count_dic = {}

        for ins in self.data:
            if ins.op.name in ["qb_alloc", "qb_dealloc"]:
                continue
            try:
                count_dic[ins.op.name] += 1
            except KeyError:
                count_dic[ins.op.name] = 1

        return count_dic

    def control(self, amount):
        return self.to_gate().control(amount)

    def compose(self, other, qubits=[], clbits=[], inplace=True):
        if inplace:
            self.append(other.to_gate(), qubits, clbits)

        else:
            res = self.copy()
            res.append(other.to_gate(), qubits, clbits)
            return res

    def bind_parameters(self, subs_dic):
        """
        Returns a QuantumCircuit where the abstract parameters in ``subs_dic`` are bound
        to their specified values.

        Parameters
        ----------
        subs_dic : dict
            A dictionary containing the abstract parameters of this QuantumCircuit as
            keys and the desired parameters as values.

        Raises
        ------
        Exception
            ``subs_dic`` did not specify a value for all abstract parameters.

        Returns
        -------
        QuantumCircuit
            The QuantumCircuit with substituted parameters.

        Examples
        --------

        We create a QuantumCircuit with some abstract parameters and bind them
        subsequently:

        >>> from qrisp import QuantumCircuit
        >>> from sympy import symbols
        >>> qc = QuantumCircuit(3)

        Create some sympy symbols and use them as abstract parameters for phase gates:

        >>> abstract_parameters = symbols("a b c")
        >>> for i in range(3): qc.p(abstract_parameters[i], i)

        Create the substitution dictionary and bind the parameters:

        >>> subs_dic = {abstract_parameters[i] : i for i in range(3)}
        >>> bound_qc = qc.bind_parameters(subs_dic)
        >>> print(bound_qc)
              ┌──────┐
        qb_0: ┤ P(0) ├
              ├──────┤
        qb_1: ┤ P(1) ├
              ├──────┤
        qb_2: ┤ P(2) ├
              └──────┘

        """

        subs_circ = self.clearcopy()

        missing_parameters = self.abstract_params - set(subs_dic.keys())
        if missing_parameters:
            raise Exception(
                "Need parameter specification for abstract parameters "
                + str(missing_parameters)
            )

        for ins in self.data:
            subs_circ.data.append(
                Instruction(ins.op.bind_parameters(subs_dic), ins.qubits, ins.clbits)
            )

        subs_circ.abstract_params = {}
        return subs_circ

    def to_latex(self, **kwargs):
        """
        Deploys the Qiskit circuit drawer to generate LaTeX output.

        Parameters
        ----------
        **kwargs : dict
            Dictionary of keyword args for Qiskits `circuit_drawer
            <https://qiskit.org/documentation/stable/0.19/stubs/qiskit.visualization.circuit_drawer.html>`_
            function.

        Returns
        -------
        string
            A string containing the latex code.

        """
        from qrisp.interface import convert_circuit

        qiskit_qc = convert_circuit(self, "qiskit", transpile=False)

        from qiskit.visualization import circuit_drawer

        return circuit_drawer(qiskit_qc, output="latex_source", **kwargs)

    def qasm(self, formatted=False, filename=None, encoding=None):
        """
        Returns the `OpenQASM <https://en.wikipedia.org/wiki/OpenQASM>`_ string of self.

        Parameters
        ----------
        formatted : bool, optional
            Return formatted Qasm string. The default is False.
        filename : string, optional
            Save Qasm to file with name ‘filename’. The default is None.
        encoding : TYPE, optional
            Optionally specify the encoding to use for the output file if filename is
            specified. By default, this is set to the system’s default encoding
            (i.e. whatever locale.getpreferredencoding() returns) and can be set to any
            valid codec or alias from stdlib’s codec module.

        Returns
        -------
        string
            The OPENQASM string.

        """
        return self.to_qiskit().qasm(formatted, filename, encoding)

    def depth(self, transpile=True):
        """
        Returns the depth of the QuantumCircuit. Note that the depth on QuantumCircuit
        which are not transpiled, might have very little correlation with the runtime.

        Parameters
        ----------
        transpile : bool, optional
            Boolean to indicate wether the QuantumCircuit should be transpiled before
            the depth is calculated. The default is True.

        Returns
        -------
        integer
            The depth of the QuantumCircuit.

        """

        from qrisp.misc import get_depth_dic

        depth_dic = get_depth_dic(self, transpile_qc=transpile)

        return max(depth_dic.values())

    def num_qubits(self):
        """
        Returns the amount of qubits.

        Returns
        -------
        int
            Amount of Qubits.

        """
        return len(self.qubits)

    # Interface for appending instructions
    # Can take either instruction or operations objects
    # Can apply multiple operations, if given the correct qubits
    # For instance if it is required to apply an x gate to qubits 4,5,6 execute
    # qc.append(XGate(), [qubit4, qubit5, qubit6])
    # If it is required to apply a cx gate to the qubit pairs (1,2), (3,4), (5,6)
    # execute qc.append(CXGate(), [[qubit1, qubit3, qubit5], [qubit2, qubit4, qubit6]])
    # If it is required to apply a cx gate to the qubit pairs (1,2), (1,3), (1,4)
    # execute qc.append(CXGate(), [qubit_1, [qubit2, qubit3, qubit4]])
    def append(self, operation_or_instruction, qubits=[], clbits=[]):
        r"""
        Method for appending Operation or Instruction objects to the QuantumCircuit.

        The parameter qubits can be an integer, a list of integers, a Qubit object or a
        list of Qubit objects. The same is valid for the clbit parameter.

        If given an Instruction object instead of an Operation, the given qubit and
        clbit parameters are ignored.

        Parameters
        ----------
        operation_or_instruction : Operation or Instruction
            The operation or instruction to be appended to the QuantumCircuit.
        qubits : integer, list[integer], Qubit, list[Qubit], optional
            The qubits on which to apply the operation. The default is [].
        clbits : integer, list[integer], Clbit, list[Clbit], optional
            The classical bits on which to apply the operation. The default is [].

        Examples
        --------

        We create a $H^{\otimes 4}$ gate and append it to every second qubit of another
        QuantumCircuit:


        >>> from qrisp import QuantumCircuit
        >>> multi_h_qc = QuantumCircuit(4, name = "multi h")
        >>> multi_h_qc.h(range(4))
        >>> multi_h = multi_h_qc.to_gate()
        >>> qc = QuantumCircuit(8)
        >>> qc.append(multi_h, [2*i for i in range(4)])
        >>> print(qc)
               ┌──────────┐
         qb_4: ┤0         ├
               │          │
         qb_5: ┤          ├
               │          │
         qb_6: ┤1         ├
               │          │
         qb_7: ┤  multi h ├
               │          │
         qb_8: ┤2         ├
               │          │
         qb_9: ┤          ├
               │          │
        qb_10: ┤3         ├
               └──────────┘
        qb_11: ────────────

        """

        # Check the type of the instruction/operation
        from qrisp.circuit import Instruction, Operation

        if self.fast_append:
            if isinstance(operation_or_instruction, Instruction):
                self.data.append(operation_or_instruction)
            else:
                if not isinstance(qubits, list):
                    raise
                self.data.append(Instruction(operation_or_instruction, qubits, clbits))
            return

        if isinstance(operation_or_instruction, Instruction):
            instruction = operation_or_instruction
            self.append(instruction.op, instruction.qubits, instruction.clbits)

            return

        elif isinstance(operation_or_instruction, Operation):
            operation = operation_or_instruction

        else:
            raise Exception(
                "Tried to append object type "
                + str(type(operation_or_instruction))
                + " which is neither Instruction nor Operation"
            )

        # Convert arguments (possibly integers) to list
        # The logic here is that the list structure gets preserved ie.
        # [[0, 1] ,2] ==> [[qubit_0, qubit_1], qubit_2]
        # unless the input is a single qubit/integer.
        # In this case we have
        # qubit_0 ==> [qubit_0]

        qubits = convert_to_qb_list(qubits, circuit=self)
        clbits = convert_to_cb_list(clbits, circuit=self)

        # Now we check which of the arguments is a list
        # For user convenience we allow to execute multiple gates at the same time
        # This comes with some restrictions where the operation to execute could be
        # ambigous.
        # When appending n gates with a single call of this function,
        # each qubit argument must either be a list of n qubits or a single qubit

        # First we check which arguments are lists
        qb_argument_is_list = []
        for i in range(len(qubits)):
            if isinstance(qubits[i], list):
                qb_argument_is_list.append(i)

        # Same with classical bits
        cb_argument_is_list = []
        for i in range(len(clbits)):
            if isinstance(clbits[i], list):
                cb_argument_is_list.append(i)

        if qb_argument_is_list + cb_argument_is_list:
            # Determine the amount of gates to be applied
            if qb_argument_is_list:
                arg_list_len = len(qubits[qb_argument_is_list[0]])
            else:
                arg_list_len = len(clbits[cb_argument_is_list[0]])

            # Check that indeed every list argument that has been given has
            # arg_list_len entries
            for arg_list_index in qb_argument_is_list:
                if len(qubits[arg_list_index]) != arg_list_len:
                    raise Exception(
                        "Don't know how to combine appending arguments "
                        + str((qubits + clbits))
                    )

            for arg_list_index in cb_argument_is_list:
                if len(clbits[arg_list_index]) != arg_list_len:
                    raise Exception(
                        "Don't know how to combine appending arguments "
                        + str((qubits + clbits))
                    )

            # Create argument constellations
            for i in range(arg_list_len):
                qubit_constellation = []
                for j in range(len(qubits)):
                    if j in qb_argument_is_list:
                        qubit_constellation.append(qubits[j][i])
                    else:
                        qubit_constellation.append(qubits[j])

                clbit_constellation = []
                for j in range(len(clbits)):
                    if j in cb_argument_is_list:
                        clbit_constellation.append(clbits[j][i])
                    else:
                        clbit_constellation.append(clbits[j])

                # Append instruction (qubit_constellation and clbit_constellation) now
                # contains no lists but only qubit/clbit arguments
                self.append(operation, qubit_constellation, clbit_constellation)

            return

        if len(qubits) != operation.num_qubits:
            raise Exception(
                f"Provided incorrect amount ({len(qubits)}) of qubits for operation "
                + str(operation.name)
            )

        if len(clbits) != operation.num_clbits:
            raise Exception(
                f"Provided incorrect amount ({len(clbits)}) of clbits for operation "
                + str(operation.name)
            )

        if len(set(qubits)) != len(qubits):
            raise Exception(
                f"Duplicate qubit arguments in {qubits} for operation {operation.name}"
            )

        # Building up the list of identifiers seems to slow down this function
        # We therefore check first if the qubit objects match and if this is not the
        # case we check if the identifiers match
        if not set(qubits).issubset(set(self.qubits)):
            op_identifiers = [qb.identifier for qb in qubits]
            qc_identifiers = [qb.identifier for qb in self.qubits]

            if not set(op_identifiers).issubset(qc_identifiers):
                raise Exception(
                    "Instruction Qubits "
                    + str(set(qubits) - set(self.qubits))
                    + " not present in circuit"
                )
            else:
                qubits = [
                    self.qubits[qc_identifiers.index(op_id)] for op_id in op_identifiers
                ]

        if len(set([cb.identifier for cb in clbits])) != len(clbits):
            raise Exception("Duplicate clbit arguments")

        if not set([cb.identifier for cb in clbits]).issubset(
            set([cb.identifier for cb in self.clbits])
        ):
            raise Exception("Instruction Clbits not present in circuit")

        # Log which abstract parameters have been added to the circuit
        try:
            self.abstract_params = self.abstract_params.union(operation.abstract_params)
        except AttributeError:
            pass

        critical_qubits = []
        perm_critical_qubits = []

        for qb in qubits:
            if qb.lock:
                critical_qubits.append(qb)
            if qb.perm_lock:
                perm_critical_qubits.append(qb)

        critical_qubits = [qb for qb in qubits if qb.lock]
        if critical_qubits:
            if critical_qubits[0].lock_message:
                raise Exception(critical_qubits[0].lock_message)
            else:
                raise Exception(
                    f"Tried to perform operation {operation.name}"
                    "on locked qubit {critical_qubits[0]}"
                )

        # Check if there are non-permeable operations on pt_locked qubits
        critical_qubits = [qb for qb in qubits if qb.perm_lock]

        if critical_qubits:
            from qrisp.uncomputation import is_permeable

            critical_qubit_indices = [qubits.index(qb) for qb in critical_qubits]
            if not is_permeable(operation, critical_qubit_indices):
                if critical_qubits[0].perm_lock_message:
                    raise Exception(critical_qubits[0].perm_lock_message)
                else:
                    raise Exception(
                        f"Tried to perform non-permeable operation {operation.name} on"
                        f" perm_locked qubit {critical_qubits[0]}"
                    )

        self.data.append(Instruction(operation, qubits, clbits))

    def run(self, shots=10000, backend=None):
        """
        Runs a QuantumCircuit on a given backend.

        Parameters
        ----------
        shots : int, optional
            The amount of shots to perform. The default is 10000.
        backend : BackendClient, optional
            The backend on which to evaluate the QuantumCircuit. The default is None.

        Returns
        -------
        dict
            The resulting counts for the given QuantumCircuit.

        Examples
        --------

        We create a GHZ QuantumCircuit and evaluate the results.

        >>> from qrisp import QuantumCircuit
        >>> qc = QuantumCircuit(5)
        >>> qc.h(0)
        >>> qc.cx(0, range(1,5))
        >>> qc.measure(range(5))
        >>> qc.run()
        {'0': 5000, '1': 5000}

        """
        if backend is None:
            from qrisp.default_backend import def_backend

            backend = def_backend

        return backend.run(self, shots)

    def statevector_array(self):
        """
        Performs a simulation of the statevector of self and returns a numpy array of
        complex numbers.

        .. note::

            Qrisps qubit ordering convention is reversed when compared to Qiskit,
            because of simulation efficiency reasons.
            As a rule of thumb you can remember:

            The statevector array of the following circuit has the amplitude 1 at the
            index ``0010 = 2``

            ::

                qb.0: ─────

                qb.1: ─────
                      ┌───┐
                qb.2: ┤ X ├
                      └───┘
                qb.3: ─────

        Returns
        -------
        numpy.ndarray
            The statevector of this circuit.

        Examples
        --------

        We create a QuantumCircuit, perform some operations and retrieve the
        statevector array.

        >>> from qrisp import QuantumCircuit
        >>> qc = QuantumCircuit(4)
        >>> qc.h(qc.qubits)
        >>> qc.z(-1)
        >>> qc.statevector_array()
        array([ 0.24999997+0.j, -0.24999997+0.j,  0.24999997+0.j, -0.24999997+0.j,
                0.24999997+0.j, -0.24999997+0.j,  0.24999997+0.j, -0.24999997+0.j,
                0.24999997+0.j, -0.24999997+0.j,  0.24999997+0.j, -0.24999997+0.j,
                0.24999997+0.j, -0.24999997+0.j,  0.24999997+0.j, -0.24999997+0.j],
              dtype=complex64)


        """
        from qrisp.simulator import statevector_sim

        return statevector_sim(self)

    def __hash__(self):
        from hashlib import sha256

        res = 0

        def hash_(x):
            temp = str(x).encode("utf-8")
            hex_value = sha256(temp).hexdigest()
            return int(hex_value, 16)

        transpiled_qc = self

        n = len(self.qubits)
        for i in range(len(transpiled_qc.data)):
            instr = transpiled_qc.data[i]

            qubit_indices = {}

            for j in range(n):
                try:
                    qubit_indices[instr.qubits.index(self.qubits[j])] = j
                except ValueError:
                    pass

            qubit_indices = [qubit_indices[j] for j in range(len(instr.qubits))]

            index_hash = hash(tuple(qubit_indices))

            params = []
            for j in range(len(instr.op.params)):
                p = hash((instr.op.params[j], i))
                params.append(p)

            param_hash = hash(tuple(params))

            if instr.op.definition:
                op_hash = hash(instr.op.definition)
            else:
                op_hash = hash(instr.op.name)

            res += hash((index_hash, param_hash, op_hash)) * (i + 1) ** 2

        res *= len(self.qubits) ** 2

        return hash(res)

    @classmethod
    def from_qiskit(self, qiskit_qc):
        """
        Class method to create QuantumCircuits from Qiskit QuantumCircuits.

        Parameters
        ----------
        qiskit_qc : Qiskit QuantumCircuit
            The Qiskit QuantumCircuit to convert.

        Returns
        -------
        QuantumCircuit
            The converted QuantumCircuit.

        Examples
        --------

        We construct a fan-out QuantumCircuit in Qiskit:

        >>> from qiskit import QuantumCircuit as QiskitQuantumCircuit
        >>> qc_2 = QiskitQuantumCircuit(4)
        >>> qc_2.cx(0, range(1,4))
        >>> print(qc_2)
        q_0: ──■────■────■──
             ┌─┴─┐  │    │
        q_1: ┤ X ├──┼────┼──
             └───┘┌─┴─┐  │
        q_2: ─────┤ X ├──┼──
                  └───┘┌─┴─┐
        q_3: ──────────┤ X ├
                       └───┘

        Note that we don't need to create a QuantumCircuit object first as this is a
        class method.

        >>> qrisp_qc_2 = QuantumCircuit.from_qiskit(qc_2)
        >>> print(qrisp_qc_2)
         qb_8: ──■────■────■──
               ┌─┴─┐  │    │
         qb_9: ┤ X ├──┼────┼──
               └───┘┌─┴─┐  │
        qb_10: ─────┤ X ├──┼──
                    └───┘┌─┴─┐
        qb_11: ──────────┤ X ├
                         └───┘

        """
        from qrisp.interface import convert_from_qiskit

        return convert_from_qiskit(qiskit_qc)

    def to_qiskit(self):
        """
        Method to convert the given QuantumCircuit to a Qiskit QuantumCircuit.

        Returns
        -------
        Qiskit QuantumCircuit
            The converted circuit.

        """
        from qrisp.interface.circuit_converter import convert_circuit

        return convert_circuit(self, target_api="qiskit", transpile=False)

    # Several methods to apply the standard operation defined in standard_operations.py

    def measure(self, qubits, clbits=None):
        """
        Instructs a measurement. If given no classical bits, the proper amount will be
        created.


        Parameters
        ----------
        qubits : Qubit
            The Qubit to be measured.
        clbits : ClBit, optional
            The Clbit to store the measurement result. The default is None.

        """
        if clbits is None:
            if isinstance(qubits, list):
                clbits = []
                for i in range(len(qubits)):
                    clbits.append(self.add_clbit())
            else:
                clbits = self.add_clbit()

        self.append(ops.Measurement(), [qubits], [clbits])

    def cx(self, qubits_0, qubits_1):
        """
        Instruct a CX-gate.

        Parameters
        ----------
        qubits_0 : Qubit
            The Qubit to control on.
        qubits_1 : Qubit
            The target Qubit.

        """
        self.append(ops.CXGate(), [qubits_0, qubits_1])

    def cy(self, qubits_0, qubits_1):
        """
        Instruct a CY-gate.

        Parameters
        ----------
        qubits_0 : Qubit
            The Qubit to control on.
        qubits_1 : Qubit
            The target Qubit.

        """
        self.append(ops.CYGate(), [qubits_0, qubits_1])

    def cz(self, qubits_0, qubits_1):
        """
        Instruct a CZ-gate.

        Parameters
        ----------
        qubits_0 : Qubit
            The Qubit to control on.
        qubits_1 : Qubit
            The target Qubit.

        """
        self.append(ops.CZGate(), [qubits_0, qubits_1])

    def h(self, qubits):
        """
        Instruct a Hadamard-gate.

        Parameters
        ----------
        qubits : Qubit
            The Qubit to apply the gate on.
        """

        self.append(ops.HGate(), [qubits])

    def x(self, qubits):
        """
        Instruct a Pauli-X-gate.

        Parameters
        ----------
        qubits : Qubit
            The Qubit to apply the gate on.
        """
        self.append(ops.XGate(), [qubits])

    def y(self, qubits):
        """
        Instruct a Pauli-Y-gate.

        Parameters
        ----------
        qubits : Qubit
            The Qubit to apply the gate on.
        """
        self.append(ops.YGate(), [qubits])

    def z(self, qubits):
        """
        Instruct a Pauli-Z-gate.

        Parameters
        ----------
        qubits : Qubit
            The Qubit to apply the gate on.
        """
        self.append(ops.ZGate(), [qubits])

    def rx(self, phi, qubits):
        """
        Instruct a parametrized RX-gate.

        Parameters
        ----------
        phi : float or sympy.Symbol
            The angle parameter.

        qubits : Qubit
            The Qubit to apply the gate on.
        """
        if phi == 0:
            return
        self.append(ops.RXGate(phi), [qubits])

    def ry(self, phi, qubits):
        """
        Instruct a parametrized RY-gate.

        Parameters
        ----------
        phi : float or sympy.Symbol
            The angle parameter.

        qubits : Qubit
            The Qubit to apply the gate on.
        """

        if phi == 0:
            return
        self.append(ops.RYGate(phi), [qubits])

    def rz(self, phi, qubits):
        """
        Instruct a parametrized RZ-gate.

        Parameters
        ----------
        phi : float or sympy.Symbol
            The angle parameter.

        qubits : Qubit
            The Qubit to apply the gate on.
        """
        if phi == 0:
            return
        self.append(ops.RZGate(phi), [qubits])

    def cp(self, phi, qubits_0, qubits_1):
        """
        Instruct a controlled phase-gate.

        Parameters
        ----------
        phi : float or sympy.Symbol
            The angle parameter.

        qubits_0 : Qubit
            The Qubit to apply the gate on.
        qubits_1 : Qubit
            The other Qubit to apply the gate on.
        """
        if phi == 0:
            return
        self.append(ops.CPGate(phi), [qubits_0, qubits_1])

    def p(self, phi, qubits):
        """
        Instruct a phase-gate.

        Parameters
        ----------
        phi : float or sympy.Symbol
            The angle parameter.

        qubits : Qubit
            The Qubit to apply the gate on.
        """
        if phi == 0:
            return
        self.append(ops.PGate(phi), [qubits])

    def rxx(self, phi, qubits_0, qubits_1):
        """
        Instruct an RXX-gate.

        Parameters
        ----------
        phi : float or sympy.Symbol
            The angle parameter.

        qubits_0 : Qubit
            The Qubit to apply the gate on.
        qubits_1 : Qubit
            The other Qubit to apply the gate on.
        """

        if phi == 0:
            return
        self.append(ops.RXXGate(phi), [qubits_0, qubits_1])

    def rzz(self, phi, qubits_0, qubits_1):
        """
        Instruct an RZZ-gate.

        Parameters
        ----------
        phi : float or sympy.Symbol
            The angle parameter.

        qubits_0 : Qubit
            The Qubit to apply the gate on.
        qubits_1 : Qubit
            The other Qubit to apply the gate on.
        """

        if phi == 0:
            return
        self.append(ops.RZZGate(phi), [qubits_0, qubits_1])

    def xxyy(self, phi, beta, qubits_0, qubits_1):
        """
        Instruct an XXYY-gate.

        Parameters
        ----------
        phi : float or sympy.Symbol
            The angle parameter.
        beta : float or sympi.Symbol
            The other angle parameter
        qubits_0 : Qubit
            The Qubit to apply the gate on.
        qubits_1 : Qubit
            The other Qubit to apply the gate on.
        """

        if phi == 0:
            return
        self.append(ops.XXYYGate(phi, beta), [qubits_0, qubits_1])

    def swap(self, qubits_0, qubits_1):
        """
        Instruct a SWAP-gate.

        Parameters
        ----------
        qubits_0 : Qubit
            The qubit to swap.
        qubits_1 : Qubit
            The other qubit to swap.

        """
        self.append(ops.SwapGate(), [qubits_0, qubits_1])

    def mcx(self, control_qubits, target_qubits, method="gray", ctrl_state=-1):
        """
        Instruct a multi-controlled X-gate.

        Parameters
        ----------
        control_qubits : list
            The list of Qubits to control on.
        target_qubits : Qubit
            The target Qubit.
        method : str, optional
            The algorithm to synthesize the mcx gate. The default is "gray".
        ctrl_state : str or int, optional
            The state on which the X gate is activated. Can be supplied as a string
            (i.e. "010110...") or an integer. The default is all ones ("11111...").


        """
        self.append(
            ops.MCXGate(len(control_qubits), ctrl_state=ctrl_state, method=method),
            control_qubits + [target_qubits],
        )

    def ccx(self, ctrl_qubit_0, ctrl_qubit_1, target_qubit, method="gray"):
        """
        Instruct a Toffoli-gate.

        Parameters
        ----------
        ctrl_qubit_0 : list
            The first control Qubit.
        ctrl_qubit_1 : Qubit
            The second control Qubit.
        target_qubit : Qubit.
            The target Qubit.
        method : str, optional
            The algorithm to synthesize the mcx gate. The default is "gray".
        """

        self.mcx([ctrl_qubit_0, ctrl_qubit_1], target_qubit, method=method)

    def crx(self, phi, qubits_0, qubits_1):
        """
        Instruct a controlled rx-gate.

        Parameters
        ----------
        phi : float or sympy.Symbol
            The angle parameter.

        qubits_0 : Qubit
            The Qubit to apply the gate on.
        qubits_1 : Qubit
            The other Qubit to apply the gate on.
        """
        if phi == 0:
            return
        self.append(ops.MCRXGate(phi, 1), [qubits_0, qubits_1])

    def t(self, qubits):
        """
        Instruct a T-gate.

        Parameters
        ----------
        qubits : Qubit
            The Qubit to apply the gate on.
        """
        self.append(ops.TGate(), [qubits])

    def t_dg(self, qubits):
        """
        Instruct a dagger T-gate.

        Parameters
        ----------
        qubits : Qubit
            The Qubit to apply the gate on.
        """
        self.append(ops.TGate().inverse(), [qubits])

    def s(self, qubits):
        """
        Instruct an S-gate.

        Parameters
        ----------
        qubits : Qubit
            The Qubit to apply the gate on.
        """
        self.append(ops.SGate(), [qubits])

    def s_dg(self, qubits):
        """
        Instruct a daggered S-gate.

        Parameters
        ----------
        qubits : Qubit
            The Qubit to apply the gate on.
        """
        self.append(ops.SGate().inverse(), [qubits])

    def sx(self, qubits):
        """
        Instruct a SX-gate.

        Parameters
        ----------
        qubits : Qubit
            The Qubit to apply the gate on.
        """
        self.append(ops.SXGate(), [qubits])

    def sx_dg(self, qubits):
        """
        Instruct a daggered SX-gate.

        Parameters
        ----------
        qubits : Qubit
            The Qubit to apply the gate on.
        """
        self.append(ops.SXGate().inverse(), [qubits])

    def barrier(self, qubits=None, clbits=None):
        """
        Instruct a Barrier onto the given Qubit. Barriers can be used as visual markers
        and compiler directives.

        Parameters
        ----------
        qubits : Qubit
            The Qubit to apply the barrier on.
        clbits : Clbit
            The Clbits to apply the barrier on.
        """

        if qubits is None:
            qubits = self.qubits

        self.append(ops.Barrier(len(qubits)), qubits)

    def reset(self, qubits):
        r"""
        Instruct a reset. This resets this Qubit into the $\ket{0}$ state regardless
        of its previous state.

        Parameters
        ----------
        qubits : Qubit
            The Qubit to reset.
        """

        self.append(ops.Reset(), [qubits])

    def u3(self, theta, phi, lam, qubits):
        r"""
        Instruct a U3-gate from given Euler angles.

        A U3 gate has the unitary:

        .. math::

            U3(\theta, \phi, \lambda) = \begin{pmatrix} \cos{(\frac{\theta}{2})}
            & -\exp{(i\lambda)}\sin{(\frac{\theta}{2})} \\
            \exp{(i\phi)} \sin{(\frac{\theta}{2})}
            & \exp{(i(\phi+\lambda))}\cos{(\frac{\theta}{2})} \end{pmatrix}

        Parameters
        ----------
        theta : float or sympy.Symbol
            The theta parameter.
        phi : float or sympy.Symbol
            The phi parameter.
        lam : float or sympy.Symbol
            The lambda parameter.
        qubits : Qubit
            The Qubit to apply the u3 gate on.

        """
        self.append(ops.u3Gate(theta, phi, lam), [qubits])

    def unitary(self, unitary_array, qubits):
        """
        Instruct a U3-gate from a given U3 matrix.

        Parameters
        ----------
        unitary_array : numpy.ndarray
            The U3 matrix to apply.
        qubits : Qubit
            The Qubit to apply the gate on.

        """

        mat = unitary_array
        coeff = 1 / np.sqrt(np.linalg.det(mat))
        gphase = -np.angle(coeff) % (2 * np.pi)
        tmp_10 = np.abs((coeff * mat[1][0]))
        tmp_00 = np.abs((coeff * mat[0][0]))
        theta = 2 * np.arctan2(tmp_10, tmp_00)
        phiplambda2 = np.angle(coeff * mat[1][1]) % (2 * np.pi)
        phimlambda2 = np.angle(coeff * mat[1][0]) % (2 * np.pi)
        phi = phiplambda2 + phimlambda2
        lam = phiplambda2 - phimlambda2

        # gphase -= (phi + lam)

        arg_max = np.argmax(np.abs(mat).flatten())
        from qrisp.simulator.unitary_management import u3matrix

        temp_u3 = u3matrix(theta, phi, lam, 0).flatten()

        gphase = (-np.angle(temp_u3[arg_max] / mat.flatten()[arg_max])) % (2 * np.pi)

        from qrisp.circuit import U3Gate
        from qrisp.simulator.unitary_management import u3matrix

        self.append(U3Gate(theta, phi, lam, global_phase=gphase), qubits)
        # self.u3(theta, phi, lam, [qubits], global_phase = gphase)

    def gphase(self, phi, qubits):
        """
        Instruct a global phase. Global phases do not directly influence the
        QuantumCircuits outcome however they can become physical if used as a base gate
        for a controlled operation.

        Parameters
        ----------
        phi : float or sympy.Symbol
            The angle parameter.
        qubits : TYPE
            The Qubit to apply the gate on.
        """

        self.append(ops.GPhaseGate(phi), [qubits])

    def id(self, qubits):
        """
        Instruct an identity gate. Identity gates are simply placeholders and have no
        effect on the quantum state.

        Parameters
        ----------

        qubits : TYPE
            The Qubit to apply the gate on.
        """

        self.append(ops.IDGate(), [qubits])


class AppendingAccelerator:
    def __enter__(self):
        self.original_appending_mode = bool(QuantumCircuit.fast_append)

        QuantumCircuit.fast_append = True

    def __exit__(self, exception_type, exception_value, traceback):
        from qrisp import QuantumCircuit

        QuantumCircuit.fast_append = self.original_appending_mode


fast_append = AppendingAccelerator


# Converts various inputs (eg. integers, qubits or quantum variables) to lists of qubit
# used in the append method of QuantumCircuit and QuantumSession
def convert_to_qb_list(input, circuit=None, top_level=True):
    from qrisp import QuantumArray
    
    if issubclass(input.__class__, Qubit):
        if top_level:
            result = [input]
        else:
            result = input
    elif isinstance(input, QuantumArray):
        result = sum([qv.reg for qv in input.flatten()], [])
        
    elif hasattr(input, "__iter__"):
        result = []
        for i in range(len(input)):
            result.append(convert_to_qb_list(input[i], circuit, top_level=False))

    elif hasattr(input, "reg"):
        result = list(input.reg)

    elif isinstance(input, int):
        if isinstance(circuit, type(None)):
            raise Exception(
                "Tried to convert integer argument to qubit without given circuit"
            )

        result = convert_to_qb_list(circuit.qubits[input], top_level=top_level)

    else:
        raise Exception("Couldn't convert type " + str(type(input)) + " to qubit list")

    return result


def convert_to_cb_list(input, circuit=None, top_level=True):
    from qrisp.circuit import Clbit

    if hasattr(input, "__iter__"):
        result = []
        for i in range(len(input)):
            result.append(convert_to_cb_list(input[i], circuit, top_level=False))

    elif isinstance(input, int):
        if isinstance(circuit, type(None)):
            raise Exception(
                "Tried to convert integer argument to qubit without given circuit"
            )

        result = convert_to_cb_list(circuit.clbits[input], top_level=top_level)

    elif issubclass(input.__class__, Clbit):
        if top_level:
            result = [input]
        else:
            result = input

    return result



"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""

from qrisp import IterationEnvironment, auto_uncompute, z, h, QuantumFloat
from qrisp.grover import diffuser
def test_iteration_env():
    
    @auto_uncompute
    def sqrt_oracle(qf):
        temp_qbool = (qf*qf == 0.25)
        z(temp_qbool)

    
    
    n = 6
    iterations = int((2**n/2)**0.5)

    qf = QuantumFloat(n-1, -1, signed = True)
    h(qf)

    with IterationEnvironment(qf.qs, iterations):
        sqrt_oracle(qf)
        diffuser(qf)
        
    mes_res = list(qf.get_measurement().keys())
    assert set([mes_res[0], mes_res[1]]) == {0.5, -0.5}
        
    qf = QuantumFloat(n-1, -1, signed = True)
    h(qf)

    with IterationEnvironment(qf.qs, iterations, precompile = True):
        sqrt_oracle(qf)
        diffuser(qf)
        
    mes_res = list(qf.get_measurement().keys())
    assert set([mes_res[0], mes_res[1]]) == {0.5, -0.5}
    


"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""


from qrisp import cx, x
from qrisp.arithmetic import create_output_qf, inpl_add


def ripple_mult(factor_1, factor_2):
    if factor_1.signed or factor_2.signed:
        raise Exception("Signed ripple multiplication currently not supported")

    s = create_output_qf([factor_1, factor_2], op="mul")

    n = factor_1.size - int(factor_1.signed)

    factor_2.exp_shift(n)

    s.init_from(factor_2)

    factor_2.exp_shift(-n)

    x(s)

    inpl_add(s, factor_2)

    for i in range(n):
        cx(factor_1[i], s)

        inpl_add(s, factor_2)

        cx(factor_1[i], s)
        factor_2.exp_shift(1)

    x(s)

    factor_2.exp_shift(-n)
    s.exp_shift(-1)

    return s



"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""

# Created by ann81984 at 04.05.2022
import pytest
import time

from qrisp.circuit import (
    multi_controlled_circuit,
    RYGate,
    RZGate,
    RXGate,
    transpile,
    HGate,
    XGate,
    PGate,
)
from qrisp import QuantumSession, QuantumVariable, QuantumCircuit, transpile, x
from qiskit.circuit.library import HGate


# Check by printing circuit
def test_controlled_gates():
    qc = QuantumCircuit(3)

    simple_case = False

    if simple_case:
        qc.x(0)
    else:
        # qc.h(0)
        # qc.cx(0,1)
        qc.cp(2.4, 0, 1)

    # print(qs)

    controlled_qc = multi_controlled_circuit(qc, 1)
    print(controlled_qc)

    ###################
    # Check by measuring circuit
    qs_0 = QuantumSession()

    qv = QuantumVariable(2, qs_0)

    if simple_case:
        qs_0.x(0)
    else:
        qs_0.h(0)
        qs_0.cx(0, 1)
        qs_0.cp(2.4, 0, 1)

    operation = qs_0.to_op()

    controlled_operation = operation.control(1)

    control_qv = QuantumVariable(1)
    qv = QuantumVariable(operation.num_qubits)

    control_on = True
    if control_on:
        x(control_qv[0])

    control_qv.qs.append(controlled_operation, control_qv.reg + qv.reg)

    tmp = qv.get_measurement()
    print(tmp)

    assert 0.4 < tmp["11"] < 0.6 and 0.4 < tmp["00"] < 0.6
    assert not (0.8 < tmp["11"] < 1 and 0.7 < tmp["00"] < 0.2)
    assert isinstance(tmp, dict)
    with pytest.raises(KeyError) as excinfo:
        bool(tmp["23"] < 0.6)
    assert "23" in str(excinfo.value)

    ###################
    start_time = time.time()

    n = 5
    test_gate = XGate().control(n)
    # test_gate = XGate().control(n)

    qs = QuantumSession()

    qv_ctrl = QuantumVariable(n, qs)
    qv_target = QuantumVariable(1, qs)

    x(qv_ctrl)

    qs.append(test_gate, qs.qubits)

    # Can be sped up to super-qiskit performance by enabling use_gray code
    # in logic_synthesis/gray_synthesis.py (this costs cnot count however)
    print("Took " + str(time.time() - start_time) + " s for gate synthesis")

    print(qv_target.get_measurement())
    tmp2 = qv_target.get_measurement()
    assert isinstance(tmp2, dict)
    assert tmp2["1"] == 1.0

    qv_0 = QuantumVariable(2)
    qv_1 = QuantumVariable(1)

    qv_0[:] = "10"
    qv_0.qs.append(XGate().control(2, ctrl_state="10"), qv_0.reg + qv_1.reg)

    assert qv_1.get_measurement() == {"1": 1.0}

    qv_0 = QuantumVariable(4)
    qv_1 = QuantumVariable(1)

    qv_0[:] = "0110"
    qv_0.qs.append(
        XGate().control(2, ctrl_state="10").control(2, ctrl_state="01"),
        qv_0.reg + qv_1.reg,
    )

    assert qv_1.get_measurement() == {"1": 1.0}



"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""

# Created by ann81984 at 06.05.2022
# import pytest


def test_qompiler():
    from qrisp import QuantumSession, QuantumVariable, x, cx

    qs = QuantumSession()

    qv_list = [QuantumVariable(1, qs=qs) for i in range(10)]

    for i in range(9):
        x(qv_list[i])
        x(qv_list[i])
        qv_list[i].delete()

    x(qv_list[-1])
    qc = qs.compile()
    assert len(qc.qubits) == 1



"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""


from qrisp.environments import QuantumEnvironment
from qrisp.circuit import Operation, ControlledOperation
from qrisp.core.session_merging_tools import recursive_qs_search, merge
from qrisp.misc import get_depth_dic

class ConjugationEnvironment(QuantumEnvironment):
    r"""
    This :ref:`QuantumEnvironment<QuantumEnvironment>` can be used for perfoming conjugated operations.
    An arbitrary unitary :math:`U \in SU(2^n)` can be conjugated by another unitary
    :math:`V \in SU(2^n)`:
        
    .. math::
        
        \text{conj}(U,V) = V^\dagger U V


    This structure appears in many quantum algorithms such as `Grover <https://arxiv.org/abs/quant-ph/9605043>`_,
    `Quantum backtracking <https://arxiv.org/abs/1509.02374>`_ or
    `Fourier arithmetic <https://arxiv.org/abs/quant-ph/0008033>`_.
    
    Using the ``ConjugationEnvironment`` not only helps to structure the code, 
    but can also grant performance advantages.
    
    This is because the controlled circuit of such a conjugation is can be
    realized by just controlling :math:`U` instead of all three operations.
    
    .. math::
        
        C\text{conj}(U,V) = V^\dagger CU V

    
    The ``ConjugationEnvironment`` can be called using the alias ``conjugate``.
    Conjugate takes the conjugation function (in our example :math:`V`) and returns
    a function that takes the arguments for the conjugation function and returns
    the corresponding ``ConjugationEnvironment``.
    For more information consult the examples section.
    
    .. note::
        
        Note that every QuantumVariable that is created by the conjugation
        function :math:`V` must be deleted/uncomputed before function conclusion.
    
    Parameters
    ----------

    conjugation_function : function
        The function performing the operation :math:`V`.
    args : iterable
        The arguments for the conjugation function.
    kwargs : dict
        The keyword arguments for the conjugation function.
        
    Examples
    --------
    
    We perform Fourier addition on a :ref:`QuantumFloat`
    
    ::
        
        from qrisp import conjugate, QuantumFloat, p, QFT
        
        def fourier_adder(qf, n):
            
            with conjugate(QFT)(qf):
                
                for i in range(qf.size):
                    p(n*np.pi*2**(i-qf.size+1), qf[i])

    >>> qf = QuantumFloat(5)
    >>> fourier_adder(qf, 3)
    >>> print(qf)
    {3: 1.0}
    >>> fourier_adder(qf, 2)
    {5: 1.0}
    
    Investigate the effects of a controlled addition:
        
    ::
        
        from qrisp import control
        
        ctrl = QuantumFloat(1)
        qf = QuantumFloat(5)
        
        with control(ctrl):
            fourier_adder(qf, 3)
    

    To see that indeed only the conjugand has been controlled we take a look
    at the circuit:
        
    >>> print(qf.qs.transpile(2))
    ctrl.0: ─────────■──────────■─────────■─────────■─────────■─────────────────
            ┌──────┐ │P(3π/16)  │         │         │         │      ┌─────────┐
      qf.0: ┤0     ├─■──────────┼─────────┼─────────┼─────────┼──────┤0        ├
            │      │            │P(3π/8)  │         │         │      │         │
      qf.1: ┤1     ├────────────■─────────┼─────────┼─────────┼──────┤1        ├
            │      │                      │P(3π/4)  │         │      │         │
      qf.2: ┤2 QFT ├──────────────────────■─────────┼─────────┼──────┤2 QFT_dg ├
            │      │                                │P(3π/2)  │      │         │
      qf.3: ┤3     ├────────────────────────────────■─────────┼──────┤3        ├
            │      │                                          │P(3π) │         │
      qf.4: ┤4     ├──────────────────────────────────────────■──────┤4        ├
            └──────┘                                                 └─────────┘


    """
    
    
    def __init__(self, conjugation_function, args, kwargs):
        
        self.conjugation_function = conjugation_function
        
        self.args = args
        
        self.kwargs = kwargs
        
        QuantumEnvironment.__init__(self)
        
    def __enter__(self):
        
        
        
        # merge(recursive_qs_search(self.args) + [self.env_qs])
        QuantumEnvironment.__enter__(self)
        
        merge(recursive_qs_search(self.args) + [self.env_qs])
        
        
        qv_set_before = set(self.env_qs.qv_list)        
        res = self.conjugation_function(*self.args, **self.kwargs)
        if qv_set_before != set(self.env_qs.qv_list):
            raise Exception("Tried to create/destroy QuantumVariables within a conjugation")
        
        self.conjugation_circ = self.env_qs.copy()
        
        self.env_qs.data = []
        
        return res
        
        
    def compile(self):
        
        temp = list(self.env_qs.data)
        self.env_qs.data = []
        
        for instr in self.conjugation_circ.data:
            if isinstance(instr, QuantumEnvironment):
                instr.compile()
            else:
                self.env_qs.append(instr)
                
        self.conjugation_circ = self.env_qs.copy()
        self.env_qs.data = []
        
        QuantumEnvironment.compile(self)
        
        content_circ = self.env_qs.copy()
        self.conjugation_circ.qubits = list(content_circ.qubits)
        
        conjugation_depth_dic = get_depth_dic(self.conjugation_circ)
        content_depth_dic = get_depth_dic(content_circ)
        
        added_depth_dic = {qb : conjugation_depth_dic[qb] + content_depth_dic[qb] for qb in content_circ.qubits}
        
        instruction_qubits = []
        
        i = 0
        while i < len(content_circ.qubits):
            
            qb = content_circ.qubits[i]
            
            if added_depth_dic[qb]:
                instruction_qubits.append(qb)
                i += 1
            else:
                content_circ.qubits.pop(i)
                self.conjugation_circ.qubits.pop(i)
        
        self.env_qs.data = temp
        
        conj_op = ConjugatedOperation(self.conjugation_circ, content_circ)
        
        self.env_qs.append(conj_op, content_circ.qubits)
        
        
class ConjugatedOperation(Operation):
    
    def __init__(self, conjugation_circ, content_circ):
        
        
        
        self.conjugation_gate = conjugation_circ.to_gate(name = "conjugator")
        self.content_gate = content_circ.to_gate(name = "conjugand")
        
        
        definition = conjugation_circ.clearcopy()
        
        definition.append(self.conjugation_gate, definition.qubits)
        definition.append(self.content_gate, definition.qubits)
        definition.append(self.conjugation_gate.inverse(), definition.qubits)
        
        Operation.__init__(self, 
                           name = "conjugation_env", 
                           definition = definition,
                           num_qubits = definition.num_qubits())
        
    def control(self, num_ctrl_qubits=1, ctrl_state=-1, method=None):
        
        controlled_conjugand = self.content_gate.control(num_ctrl_qubits=num_ctrl_qubits, ctrl_state=ctrl_state, method=None)
        
        res = type(controlled_conjugand)(self, num_ctrl_qubits=num_ctrl_qubits, ctrl_state=ctrl_state, method=method)
        
        res.definition.data = []
        
        res.definition.append(self.conjugation_gate, self.definition.qubits)
        res.definition.append(controlled_conjugand, res.definition.qubits[:num_ctrl_qubits] + self.definition.qubits)
        res.definition.append(self.conjugation_gate.inverse(), self.definition.qubits)
        
        return res
        
        
def conjugate(conjugation_function):

     def conjugation_env_creator(*args, **kwargs):
         
         return ConjugationEnvironment(conjugation_function, args, kwargs)
     
     return conjugation_env_creator
        
        
            
    
        


"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""


class Instruction:
    """
    This class combines Operation objects with their operands (ie. qubits and classical
    bits). The data attribut of the QuantumCircuit class consists of a list of
    Instructions.

    Instructions can be added to QuantumCircuits using the
    `append <qrisp.QuantumCircuit.append>` method without any
    qubit/ classical bit arguments.

    Parameters
    ----------
    op : Operation
        The Operation object.
    qubits : list[Qubit], optional
        The list of Qubits on which op should take place. The default is [].
    clbits : list[Clbit], optional
        The list of Clbits on which op should take place. The default is [].


    Examples
    --------

    We create two Instruction objects, merge them and append the result to a
    QuantumCircuit.

    >>> from qrisp import Instruction, QuantumCircuit, CXGate, XGate
    >>> qc = QuantumCircuit(2)
    >>> ins_0 = Instruction(XGate(), [qc.qubits[0]])
    >>> ins_1 = Instruction(CXGate(), qc.qubits)
    >>> merged_ins = ins_0.merge(ins_1)
    >>> qc.append(merged_ins)
    >>> qc.measure(qc.qubits[1])
    >>> qc.run()
    {'1': 10000}
    >>> print(qc.transpile())
          ┌───┐
    qb_0: ┤ X ├──■─────
          └───┘┌─┴─┐┌─┐
    qb_1: ─────┤ X ├┤M├
               └───┘└╥┘
    cb_0: ═══════════╩═

    """

    def __init__(self, op, qubits=[], clbits=[]):
        self.op = op
        self.qubits = qubits
        self.clbits = clbits

    def merge(self, other):
        """
        Merges two instructions into one.

        Parameters
        ----------
        other : Instruction
            The second instruction.

        Returns
        -------
        res : Instruction
            The merged instruction (self is executed first).

        """

        from qrisp.circuit import QuantumCircuit

        qubit_list = list(set(self.qubits + other.qubits))
        clbit_list = list(set(self.clbits + other.clbits))

        qubit_list.sort(key=lambda x: x.identifier)
        clbit_list.sort(key=lambda x: x.identifier)
        qc = QuantumCircuit()

        for i in range(len(qubit_list)):
            qc.add_qubit(qubit_list[i])

        for i in range(len(clbit_list)):
            qc.add_clbit(clbit_list[i])

        qc.data = [self, other]

        res = Instruction(qc.to_op(), qubit_list, clbit_list)
        return res

    def copy(self):
        """
        Returns a copy of the Instruction.

        Returns
        -------
        Instruction
            The copied Instruction.

        """
        return Instruction(self.op.copy(), list(self.qubits), list(self.clbits))

    def inverse(self):
        res = self.copy()
        res.op = res.op.inverse()

        return res

    def __str__(self):
        if len(self.clbits):
            return (
                self.op.name
                + "("
                + str(self.qubits)[1:-1]
                + ", "
                + str(self.clbits)[1:-1]
                + ")"
            )
        else:
            return self.op.name + "(" + str(self.qubits)[1:-1] + ")"

    def __repr__(self):
        return self.__str__()



"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""


class Qubit:
    """
    This class describes qubits. Qubits are created by supplying the identifier string.

    Attributes
    ----------
    identifier : str
        A string to identify the Qubit.

    Examples
    --------

    We create a Qubit and add it to a :ref:`QuantumCircuit`:

    >>> from qrisp import QuantumCircuit, Qubit
    >>> qb = Qubit("alphonse")
    >>> qc = QuantumCircuit()
    >>> qc.add_qubit(qb)
    >>> qc.x(qb)
    >>> print(qc)
              ┌───┐
    alphonse: ┤ X ├
              └───┘


    """

    def __init__(self, identifier):
        self.identifier = identifier
        self.hash_value = id(self)
        self.lock = False
        self.perm_lock = False

    def __str__(self):
        return self.identifier

    def __repr__(self):
        return "Qubit(" + self.identifier + ")"

    def __hash__(self):
        return self.hash_value

    def __eq__(self, other):
        return self.hash_value == other.hash_value



"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""

import numpy as np

from qrisp import QuantumBool, QuantumChar, QuantumFloat, QuantumSession, merge
from qrisp.circuit import QuantumCircuit, QubitAlloc, QubitDealloc, transpile
from qrisp.quantum_network import QuantumNetworkClient


class QuantumNetworkSession(QuantumSession):
    """
    This class allows using Qrisps high-level programming interface to interact with
    QuantumNetworks. As an inheritor of the QuantumSession class, we can create
    QuantumVariables with it which can then be sent to other participators of the
    QuantumNetwork.

    Note that get_measurement calls are also realized by the single-shot simulator. For
    more information on this check the documentation of the run method of the
    QuantumNetworkClient class.

    Parameters
    ----------
    name : str
        A nickname such as "Alice" or "Bob", which is used to direct the communication
        of other users to you.
    socket_ip : str
        The IP of the QuantumNetworkServer to connect to.
    port : int
        The port of the QuantumNetworkServer to connect to. The default is 7070.

    Examples
    --------

    We create a QuantumNetworkServer and connect a QuantumNetworkSession to it.

    >>> from qrisp.quantum_network import QuantumNetworkServer, QuantumNetworkSession
    >>> example_server = QuantumNetworkServer(socket_ip = "127.0.0.1", background = True)
    >>> example_server.start()
    >>> alice_session = QuantumNetworkSession(socket_ip = "127.0.0.1", port = 7070, name = "alice")

    Now we can create QuantumVariables as we are used to

    >>> from qrisp import QuantumFloat
    >>> qf = QuantumFloat(4, 5, qs = alice_session)


    """

    def __init__(self, name, socket_ip, port=None):
        qn_client = QuantumNetworkClient(name=name, socket_ip=socket_ip, port=port)
        self.inbox = []

        super().__init__(backend=qn_client)

    def send_qv(self, qv, recipient):
        """
        Sends a :ref:`QuantumVariable` to another participant of the network.

        Parameters
        ----------
        qv : QuantumVariable
            The QuantumVariable to send.
        recipient : str
            The name of the recipient.

        Examples
        --------

        We create a QuantumNetworkServer and connect a QuantumNetworkSession to it.

        >>> from qrisp.quantum_network import QuantumNetworkServer, QuantumNetworkSession
        >>> example_server = QuantumNetworkServer(socket_ip = "127.0.0.1", background = True)
        >>> example_server.start()
        >>> alice_session = QuantumNetworkSession(socket_ip = "127.0.0.1", port = 7070, name = "alice")
        >>> bob_session = QuantumNetworkSession(socket_ip = "127.0.0.1", port = 7070, name = "bob")

        We create a :ref:`QuantumFloat` and perform some arithmetic

        >>> from qrisp import QuantumFloat
        >>> qf = QuantumFloat(4, qs = alice_session)
        >>> qf += 3
        >>> alice_session.send_qv(qf, "bob")


        """

        if not qv.qs == self:
            merge(self, qv.qs)

        self.backend.run(self)

        qv_qubits = list(qv.reg)

        type_information = {}

        if isinstance(qv, QuantumFloat):
            type_information["qtype"] = "qf"
            type_information["msize"] = str(qv.msize)
            type_information["exponent"] = str(qv.exponent)
            type_information["signed"] = str(int(qv.signed))
        elif isinstance(qv, QuantumChar):
            type_information["qtype"] = "qc"
            type_information["nisq_char"] = str(int(qv.nisq_char))
        elif isinstance(qv, QuantumBool):
            type_information["qtype"] = "qb"

        self.backend.send_qubits(recipient, qv_qubits, str(type_information))

        qv.delete()

        for i in range(len(qv_qubits)):
            self.qubits.remove(qv_qubits[i])

        self.data = []

    def update(self):
        """
        This updates the inbox attribute and returns the received QuantumVariables.

        Returns
        -------
        received_qvs : list[QuantumVariable]
            The QuantumVariables which have been received since the last update.

        Examples
        --------

        We assume that the commands from the example of the
        :meth:`send_qv <qrisp.quantum_network.QuantumNetworkSession.send_qv>` method
        have been executed.

        >>> inbox = bob_session.update()
        >>> print(inbox[0])
        {3.0: 1.0}

        """

        from qrisp import QuantumVariable

        self.backend.run(self)

        updated_inbox = self.backend.inbox()

        received_qvs = []
        starting_index = int(len(self.inbox))
        from ast import literal_eval

        for i in range(starting_index, len(updated_inbox)):
            for qb in updated_inbox[i][0]:
                self.add_qubit(qb)

            try:
                type_information = literal_eval(updated_inbox[i][1])
            except:
                type_information = {}

            new_qv = QuantumVariable(0, qs=self)

            new_qv.extend(len(updated_inbox[i][0]), proposal=updated_inbox[i][0])

            if len(type_information):
                if type_information["qtype"] == "qf":
                    new_qv.__class__ = QuantumFloat
                    new_qv.msize = int(type_information["msize"])
                    new_qv.exponent = int(type_information["exponent"])
                    new_qv.signed = bool(int(type_information["signed"]))

                if type_information["qtype"] == "qc":
                    new_qv.__class__ = QuantumChar
                    new_qv.nisq_char = bool(int(type_information["nisq_char"]))

                elif type_information["qtype"] == "qb":
                    new_qv.__class__ = QuantumBool

            received_qvs.append(new_qv)

        self.inbox += received_qvs
        return received_qvs



"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""

import numpy as np

# Checks if a gate is permeable on a set of qubits
# Permeable means that the gate has the form
# U = \sum_{x = 0}^{2^n} |x><x| U_x
# where the braket expression represents the qubits
# of the list "qubit_indices"
# {U_x | x <= 2^n} is a set of operations on the remaining qubits

# Applying a gate which is permeable on a register R to a computational basis state in R
# leaves the state of R invariant i.e. if R is in state |x>
# U |x>|y> = |x> U_x |y>
# this property will for instance be used to make sure that uncomputations,
# which depend on the state of R indeed stil recieve the original state of R
# when the uncomputation is performed. An error will be thrown if in the meantime,
# there have been non-permeable operations on R

# But what is the general strategy in determining this property?
# To understand this, note that if we order the computational basis vectors,
# such that all the qubits which are supposed to be checked for permeability are at the
# end.

# In this basis, the unitary is block diagonal, where the blocks on the diagonal are the
# U_x matrices, we check the block diagonality by first calculating the block matrices
# and then create a boolean 2^n x 2^n matrix, where the entry is False if the
# corresponding block matrix is 0


# We then check the diagonality of this matrix and return True if so
def is_permeable(gate, qubit_indices):
    for i in qubit_indices:
        if gate.permeability[i] is False:
            return False
        elif gate.permeability[i] is None:
            break
    else:
        return True

    from qrisp.simulator import calc_embedded_unitary

    if gate.definition:
        qc = gate.definition
        for instr in qc.data:
            relevant_qubits = set([qc.qubits[k] for k in qubit_indices]).intersection(
                instr.qubits
            )

            if not relevant_qubits:
                continue

            if instr.op.name == "measure":
                return False

            local_qubit_indices = [instr.qubits.index(qb) for qb in relevant_qubits]

            if not is_permeable(instr.op, local_qubit_indices):
                break

        else:
            for i in qubit_indices:
                gate.permeability[i] = True

            return True

    # Create qubit order
    qubit_order = []
    for i in range(gate.num_qubits):
        if i not in qubit_indices:
            qubit_order.append(i)

    # The qubits which are to be checked come first
    qubit_order = qubit_indices + qubit_order

    # Invert permutation
    qubit_order = [qubit_order.index(i) for i in range(len(qubit_order))]

    # Calculate unitary
    unitary = calc_embedded_unitary(gate, gate.num_qubits, qubit_order)

    if not isinstance(unitary, np.ndarray):
        unitary = unitary.to_array()

    # Calculate boolean matrix
    bbm = get_boolean_block_matrix(unitary, 2 ** (gate.num_qubits - len(qubit_indices)))
    # Check non-zero off diagonal elements
    off_diagonal_non_zero_count = np.count_nonzero(bbm - np.diag(np.diagonal(bbm)))
    # Return result

    result = not bool(off_diagonal_non_zero_count)

    if result:
        for i in qubit_indices:
            gate.permeability[i] = True

    return result


# Tests if a gate is qfree where qfree is the property that is introduced in SILQ
# meaning that an operation maps any computational basis state to a computational basis
# state.
def is_qfree(gate):
    if gate.is_qfree is not None:
        return gate.is_qfree

    if gate.definition:
        for instr in gate.definition.data:
            if not is_qfree(instr.op):
                break
        else:
            gate.is_qfree = True
            return True

    unitary = np.around(gate.get_unitary(), 5)
    # Check if the unitary has only one non-zero entry in each line
    for i in range(unitary.shape[0]):
        if np.count_nonzero(unitary[i, :]) != 1:
            gate.is_qfree = False
            return False

    gate.is_qfree = True
    return True


# Splits the given matrix into blocks of size block_size
# Returns an object array where each entry contains the corresponding block
def get_block_matrix(matrix, block_size):
    matrix = np.array(matrix)
    if matrix.shape[0] != matrix.shape[1]:
        raise Exception("Tried to decompose non-square blockmatrix")

    if matrix.shape[0] % block_size:
        raise Exception("Matrix shape is no integer multiple of block size")

    block_amount = int(matrix.shape[0] / block_size)
    result = np.zeros((block_amount, block_amount), dtype="object")

    for i in range(block_amount):
        for j in range(block_amount):
            result[i, j] = matrix[
                i * block_size : (i + 1) * block_size,
                j * block_size : (j + 1) * block_size,
            ]

    return result


# Recieves a matrix which is turned into a block matrix and then
# returns a matrix of boolean entries where the entry is False, if
# the corresponding block is the 0 matrix and True otherwise
def get_boolean_block_matrix(matrix, block_size):
    block_matrix = get_block_matrix(matrix, block_size)

    block_amount = int(matrix.shape[0] / block_size)

    result = np.zeros((block_amount, block_amount))

    for i in range(block_amount):
        for j in range(block_amount):
            if np.linalg.norm(block_matrix[i, j]) > 1e-4:
                result[i, j] = 1

    return result



"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""


import sys

from qrisp.interface.circuit_converter import convert_circuit
from qrisp.interface.thrift_interface.codegen import BackendService


class BackendClient(BackendService.Client):
    """
    This object allows connecting to Qrisp backend servers.

    Parameters
    ----------
    socket_ip : string
        The IP address of the socket of the target server.
    port : int
        The port on which the server is listening.


    Examples
    --------

    We assume that the example from BackendServer has been executed in the same console.

    >>> from qrisp.interface import BackendClient
    >>> example_backend = BackendClient(socket_ip = "127.0.0.1", port = 8080)
    >>> from qrisp import QuantumCircuit
    >>> qc = QuantumCircuit(2)
    >>> qc.h(0)
    >>> qc.cx(0,1)
    >>> qc.measure(qc.qubits)
    >>> example_backend.run(qc, shots = 1000, token = "lorem ipsum")
    lorem ipsum
    {'0 0': 464, '1 1': 536}



    """

    def __init__(self, socket_ip, port=None):
        from thrift.protocol import TBinaryProtocol
        from thrift.transport import TSocket, TTransport

        if port is None:
            port = 9010
        # Create the transport for the User Interface to the server
        self.transport = TSocket.TSocket(socket_ip, port)
        # Buffering is critical. Raw sockets are very slow
        self.transport = TTransport.TBufferedTransport(self.transport)
        # Wrap in a protocol
        protocol = TBinaryProtocol.TBinaryProtocol(self.transport)

        self.socket_ip = socket_ip

        # Create a client to use the protocol encoder
        # client = BackendService.Client(protocol)

        super().__init__(protocol)

        # Connect!
        self.transport.open()

    # Destructor closes the transport
    def __del__(self):
        self.transport.close()

    def run(self, qc, shots, token=""):
        """
        Executes the ``run_func`` of the server.

        Parameters
        ----------
        qc : QuantumCircuit
            The QuantumCircuit to execute.
        shots : int
            The amount of shots.
        token : str, optional
            A token which can be used for backend execution parameter specification.
            The default is "".

        Returns
        -------
        dict
            A dictionary representing the counts where the keys are bitstrings and the
            values are integers.

        """

        converted_circuit = convert_circuit(qc, "thrift", transpile=False)

        return super().run(converted_circuit, shots, token)

    def ping(self):
        """
        Executes the ``ping_func`` on the server.

        Returns
        -------
        BackendStatus object
            An object which contains general information about the backend.
        """

        return super().ping()



"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""

import sympy as sp
from qrisp.logic_synthesis.truth_tables import TruthTable, synth_poly


def pprm_synth(input_var, output_var, tt, qb_nr, phase_tolerant=False):
    from qrisp import mcx
    from qrisp.arithmetic.poly_tools import (
        expr_to_list,
    )

    qs = input_var.qs
    output_qubit = output_var.reg[qb_nr]
    expr = synth_poly(tt, column=qb_nr)
    # print(expr)
    args = expr_to_list(expr)
    args_len = [len(i) for i in args]

    for element in range(len(args_len)):
        if args_len[element] > 1:
            from sympy.core.numbers import One

            if isinstance(args[element][0], One):
                args[element].pop(0)

            # input_qubits = [(-int(symb.name[1:])-1)%input_var.size for symb in args[element]]  # noqa
            input_qubits = [int(symb.name[1:]) for symb in args[element]]
            input_qubits = [input_var.reg[i] for i in input_qubits]

            product = 1
            for i in range(len(args[element])):
                product = product * (args[element][i])

            mul_tt = TruthTable(product)

            if len(product.args) == 0:
                if isinstance(product, sp.Symbol):
                    qs.cx(input_qubits[0], output_qubit)
                else:
                    qs.x(output_qubit)
            else:
                if phase_tolerant:
                    qs.append(
                        mul_tt.gate_synth(method="gray_pt"),
                        input_qubits + [output_qubit],
                    )
                else:
                    mcx(input_qubits, output_qubit, method="gray")
                    # qs.append(mul_tt.gate_synth(method = "gray"),input_qubits+[output_qubit])  # noqa
        else:
            if isinstance(args[element][0], sp.core.symbol.Symbol):
                arg_qubit = (-int(args[element][0].name[1:]) - 1) % input_var.size

                arg_qubit = input_var.reg[arg_qubit]
                qs.cx(arg_qubit, output_qubit)
            if isinstance(args[element][0], sp.core.numbers.One):
                qs.x(output_qubit)


def pprm(input_var, output_var, tt, phase_tolerant=False):
    for column in range(tt.shape[1]):
        pprm_synth(input_var, output_var, tt, column, phase_tolerant)



"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""


def test_quantum_arithmetic():
    import numpy as np
    from qrisp import h, multi_measurement, q_div, QuantumFloat, QuantumBool

    def test_arithmetic_helper(qf_0, qf_1, operation):
        qf_0 = qf_0.duplicate()
        qf_1 = qf_1.duplicate()

        h(qf_0)
        h(qf_1)

        if operation == "add":
            qf_res = qf_0 + qf_1
        elif operation == "sub":
            qf_res = qf_0 - qf_1
        elif operation == "mul":
            qf_res = qf_0 * qf_1
        elif operation == "div":
            qf_res = q_div(qf_0, qf_1, prec=3)

        mes_res = multi_measurement([qf_0, qf_1, qf_res])

        for a, b, c in mes_res.keys():
            if operation == "add":
                assert a + b == c
            elif operation == "sub":
                assert a - b == c
            elif operation == "mul":
                assert a * b == c
            elif operation == "div":
                if b == 0:
                    continue
                assert abs(a / b - c) < 2 ** (-3)

        statevector = qf_res.qs.statevector("array")
        angles = np.angle(
            statevector[
                np.abs(statevector) > 1 / 2 ** ((qf_0.size + qf_1.size) / 2 + 1)
            ]
        )

        # Test correct phase behavior
        assert np.sum(np.abs(angles)) < 0.1

    a = QuantumFloat(3, -1, signed=True)
    b = QuantumFloat(5, 1, signed=False)

    for operation in ["add", "sub", "mul", "div"]:
        print(operation)
        test_arithmetic_helper(a, b, operation)

    a = QuantumFloat(2, 1, signed=True)
    b = QuantumFloat(3, 0, signed=True)

    for operation in ["add", "sub", "mul", "div"]:
        test_arithmetic_helper(a, b, operation)

    # Test tensordot
    from qrisp import QuantumFloat, QuantumArray, tensordot, OutcomeArray

    qf = QuantumFloat(3, 0, signed=False)
    q_tensor_0 = QuantumArray(qf, shape=(4, 4))
    q_tensor_0[:] = np.eye(4)
    q_tensor_0 = q_tensor_0.reshape(2, 2, 2, 2)
    q_tensor_1 = QuantumArray(qf, shape=(2, 2))
    q_tensor_1[:] = [[0, 1], [1, 0]]
    res = tensordot(q_tensor_0, q_tensor_1, (-1, 0))

    assert res.get_measurement() == {
        OutcomeArray(
            [[[[0, 1], [0, 0]], [[1, 0], [0, 0]]], [[[0, 0], [0, 1]], [[0, 0], [1, 0]]]]
        ): 1.0
    }

    # Test inplace multiplication
    b = QuantumFloat(5, signed=True)
    b[:] = 3
    h(b[0])
    h(b[-1])
    b *= -4
    assert b.get_measurement() == {116: 0.25, 120: 0.25, -12: 0.25, -8: 0.25}
    
    qf = QuantumFloat(4, signed = True)
    qf[:] = 1

    qbl = QuantumBool()

    h(qbl)

    with qbl:
        qf *= -3*0.125
    
    assert qf.get_measurement() == {1.0: 0.5, -0.375: 0.5}

    qf = QuantumFloat(4)
    h(qf)
    assert qf.get_ev() == 7.5
            


"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""


from qrisp.circuit import QuantumCircuit, Qubit, XGate


# This function takes a circuit and turns it into it's controlled version
def multi_controlled_circuit(
    input_circuit, control_amount=1, ctrl_state=-1, method="gray"
):
    # Create result circuit
    controlled_circuit = QuantumCircuit()

    # Add control qubits
    for i in range(control_amount):
        controlled_circuit.add_qubit(Qubit(identifier="ctrl_qb_" + str(i)))

    # Set alias for qubits
    control_qubits = list(controlled_circuit.qubits)

    # Add circuit qubits
    for qb in input_circuit.qubits:
        controlled_circuit.add_qubit(qb)

    # Apply controlled instructions
    for ins in input_circuit.data:
        if ins.op.name in ["qb_alloc", "qb_dealloc"]:
            # raise
            controlled_circuit.append(ins.op, ins.qubits)
            continue
        controlled_circuit.append(
            ins.op.control(control_amount, method=method, ctrl_state=ctrl_state),
            control_qubits + ins.qubits,
        )

    return controlled_circuit


# This function takes an U3Gate object and turns it into it's controlled version
def multi_controlled_u3_circ(u3_gate, control_amount, ctrl_state, method="gray"):
    from qrisp.misc.multi_cx import multi_cx
    from qrisp.logic_synthesis import gray_phase_synth_qb_list

    qc = QuantumCircuit(control_amount + 1)

    # Apply control state specification
    for i in range(len(ctrl_state)):
        if ctrl_state[i] == "0":
            qc.x(i)

    # If the U3 gate is an rx , ry, rz or p gate, we can use gray phase synthesis on the
    # target qubit and wrap this in the corresponding gates
    # (for instance H for rx because RX = H RZ H)
    if u3_gate.name in ["rz", "p"]:
        # Synthesize phases using gray synthesis
        gray_phase_synth_qb_list(
            qc, qc.qubits, (2 ** (control_amount + 1) - 1) * [0] + [u3_gate.params[0]]
        )

    elif u3_gate.name == "rz":
        # Make sure to account for the possible global phase of rz
        gray_phase_synth_qb_list(
            qc,
            qc.qubits,
            (2 ** (control_amount + 1) - 2) * [0]
            + [-u3_gate.params[0] / 2, u3_gate.params[0] / 2],
        )

        qc.global_phase -= u3_gate.params[0]

    elif u3_gate.name == "rx":
        # Same thing as with rz but now we use RX = H RZ H
        qc.h(-1)
        gray_phase_synth_qb_list(
            qc,
            qc.qubits,
            (2 ** (control_amount + 1) - 2) * [0]
            + [-u3_gate.theta / 2, u3_gate.theta / 2],
        )
        qc.h(-1)

    elif u3_gate.name == "ry":
        # Now we use RY = S RX S_DG
        qc.s(-1)
        qc.h(-1)
        gray_phase_synth_qb_list(
            qc,
            qc.qubits,
            (2 ** (control_amount + 1) - 2) * [0]
            + [u3_gate.theta / 2, -u3_gate.theta / 2],
        )
        qc.h(-1)
        qc.s_dg(-1)

    elif u3_gate.phi == 0 and u3_gate.lam == 0 and u3_gate.theta == 0:
        pass

    # Treat pauli gates
    elif u3_gate.name == "y":
        qc.s(-1)
        qc.append(XGate().control(control_amount, method=method), qc.qubits)
        qc.s_dg(-1)

    elif u3_gate.name == "z":
        qc.h(-1)
        qc.append(XGate().control(control_amount, method=method), qc.qubits)
        qc.h(-1)

    elif u3_gate.name == "x":
        qc.append(multi_cx(control_amount, method), qc.qubits)

    # Treat general U3Gates
    else:
        # Algorithm based on https://arxiv.org/pdf/quant-ph/9503016.pdf
        alpha = u3_gate.phi
        theta = -u3_gate.theta
        beta = u3_gate.lam

        A = QuantumCircuit(1)

        A.p(alpha, 0)
        A.ry(theta / 2, 0)

        B = QuantumCircuit(1)

        B.ry(-theta / 2, 0)
        B.p(-(alpha + beta) / 2, 0)

        C = QuantumCircuit(1)

        C.p((beta - alpha) / 2, 0)

        # Treat global phases and the fact that X P(phi) X = exp(2 phi) P(-phi)
        if method not in ["gray_pt", "gray_pt_inv"]:
            control_phase = -u3_gate.global_phase / 2 - (alpha + beta)
            gray_phase_synth_qb_list(
                qc,
                qc.qubits[:-1],
                (2 ** (control_amount) - 1) * [0] + [-control_phase / 2],
            )

        qc.append(A.to_gate("A"), qc.qubits[-1])

        # To perform the controlled x gate, we can use the phase tolerant algorithm
        mcx_gate = XGate().control(control_amount, method = "gray_pt")

        qc.append(mcx_gate, qc.qubits)

        qc.append(B.to_gate("B"), qc.qubits[-1])

        qc.append(mcx_gate.inverse(), qc.qubits)

        qc.append(C.to_gate("C"), qc.qubits[-1])

    # Un-apply control state specification
    for i in range(len(ctrl_state)):
        if ctrl_state[i] == "0":
            qc.x(i)

    return qc


def multi_controlled_gray_circ(gray_gate, control_amount, ctrl_state):
    from qrisp.logic_synthesis.gray_synthesis import gray_synth_gate

    target_phases_old = gray_gate.target_phases

    ctrl_state = int(ctrl_state, 2)

    target_phases_new = []

    for i in range(2**control_amount):
        if i != ctrl_state:
            target_phases_new += [0] * len(target_phases_old)
        else:
            target_phases_new += target_phases_old

    new_gray_gate = gray_synth_gate(target_phases_new)

    return new_gray_gate.definition, target_phases_new



"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""

# Created by ann81984 at 07.05.2022
import numpy as np
from qrisp import *

from qrisp import QuantumVariable, mcx, cx, QuantumEnvironment, x


def test_uncomputation_example():
    verify[0] = 1

    qf1 = QuantumFloat(3, name="a")

    qf2 = QuantumFloat(3, name="b")
    qf1[:] = 4

    h(qf1[0])

    qf2[:] = 5

    res = qf1 * qf2

    temp = res.duplicate()

    qbool = QuantumBool()
    h(qbool)

    with qbool == True:
        cx(res, temp)

    res.uncompute()

    assert multi_measurement([temp, qbool]) == {
        (0.0, False): 0.5,
        (20.0, True): 0.25,
        (25.0, True): 0.25,
    }

    # ------------------
    print("Test 1 passed")

    @auto_uncompute
    def sqrt_oracle(qf):
        z(qf * qf == 0.25)

    qf = QuantumFloat(3, -1, signed=False)
    n = qf.size
    iterations = int((2**n) ** 0.5)
    h(qf)

    from qrisp.grover import diffuser

    for i in range(iterations):
        sqrt_oracle(qf)
        diffuser(qf)
        print(qf)

    assert qf.get_measurement() == {
        0.5: 0.9453,
        0.0: 0.0078,
        1.0: 0.0078,
        1.5: 0.0078,
        2.0: 0.0078,
        2.5: 0.0078,
        3.0: 0.0078,
        3.5: 0.0078,
    }

    # ---------
    print("Test 2 passed")

    city_amount = 4

    distance_matrix = (
        np.array(
            [
                [0, 0.25, 0.125, 0.5],
                [0.25, 0, 0.625, 0.375],
                [0.125, 0.625, 0, 0.75],
                [0.5, 0.375, 0.75, 0],
            ]
        )
        / 4
    )

    def calc_perm_travel_distance(permutation, precision):
        # A QuantumFloat with n qubits and exponent -n
        # can represent values between 0 and 1
        res = QuantumFloat(precision, -precision)

        # Fill QuantumDictionary
        qd = QuantumDictionary(return_type=res)
        for i in range(city_amount):
            for j in range(city_amount):
                qd[(i, j)] = distance_matrix[i, j]

        # Evaluate result
        for i in range(4):
            trip_distance = qd[permutation[i], permutation[(i + 1) % city_amount]]
            res += trip_distance

            trip_distance.uncompute()

        return res

    perm = QuantumArray(QuantumFloat(2))
    perm[:] = [2, 1, 0, 3]

    res = calc_perm_travel_distance(perm, 5)

    assert res.get_measurement() == {0.53125: 1.0}

    assert len(res.qs.compile().qubits) == 18

    # --------
    print("Test 3 passed")

    def triple_AND(a, b, c):
        local_quantum_bool = QuantumBool(name="local")
        result = QuantumBool(name="result")

        mcx([a, b], local_quantum_bool)
        mcx([local_quantum_bool, c], result)

        local_quantum_bool.uncompute()

        return result

    a = QuantumBool(name="a")
    b = QuantumBool(name="b")
    c = QuantumBool(name="c")

    h(a)
    b[:] = True
    c[:] = True

    result = triple_AND(a, b, c)

    assert result.get_measurement() == {False: 0.5, True: 0.5}

    # --------

    @auto_uncompute
    def triple_AND(a, b, c):
        local_quantum_bool = QuantumBool(name="local")
        result = QuantumBool(name="result")

        mcx([a, b], local_quantum_bool)
        mcx([local_quantum_bool, c], result)

        return result

    a = QuantumBool(name="a")
    b = QuantumBool(name="b")
    c = QuantumBool(name="c")

    h(a)
    b[:] = True
    c[:] = True

    result = triple_AND(a, b, c)

    assert result.get_measurement() == {False: 0.5, True: 0.5}

    print("Test 4 passed")

    # --------

    # Test uncomputation/environment interaction

    error_thrown = False

    d = QuantumVariable(1, name="d")
    cond_1 = QuantumBool(name="cond_1")
    cond_2 = QuantumBool(name="cond_2")
    cond_3 = QuantumBool(name="cond_3")

    with cond_1:
        a = QuantumVariable(1, name="a")
        b = QuantumVariable(1, name="b")
        # x(a)
        x(b)
        with cond_2:
            x(a)
            x(d)
            with cond_3:
                x(b)
            try:
                a.uncompute()
            except:
                error_thrown = True

        b.uncompute()

    assert error_thrown

    print("Test 5 passed")

    # Test recomputation

    qv_0 = QuantumVariable(1)
    qv_1 = QuantumVariable(1)
    qv_2 = QuantumVariable(1)

    cx(qv_0, qv_1)
    cx(qv_1, qv_2)

    qv_1.uncompute(recompute=True)

    qv_2.uncompute()
    print(qv_1.qs)

    qv_0 = QuantumVariable(1)
    qv_1 = QuantumVariable(1)
    qv_2 = QuantumVariable(1)
    qv_3 = QuantumVariable(1)

    cx(qv_0, qv_1)
    cx(qv_1, qv_2)
    cx(qv_1, qv_3)
    x(qv_2)

    qv_1.uncompute(recompute=True)
    qv_2.uncompute()
    qv_3.uncompute()
    assert qv_1.qs.cnot_count() == 10

    verify[0] = 0



"""
/*********************************************************************
* Copyright (c) 2023 the Qrisp Authors
*
* This program and the accompanying materials are made
* available under the terms of the Eclipse Public License 2.0
* which is available at https://www.eclipse.org/legal/epl-2.0/
*
* SPDX-License-Identifier: EPL-2.0
**********************************************************************/
"""



"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""


from qrisp import Qubit
from qrisp.interface import PortableQubit
from qrisp.interface.circuit_converter import convert_circuit
from qrisp.quantum_network.interface.codegen import QuantumNetworkService
from qrisp.quantum_network.interface.codegen.ttypes import Message


class QuantumNetworkClient(QuantumNetworkService.Client):
    """
    This class allows connecting to Qrisp quantum networks.

    Parameters
    ----------
    name : str
        A nickname such as "Alice" or "Bob", which is used to direct the communication
        of other users to you.
    socket_ip : str
        The IP adress of the QuantumNetworkServer.
    port : int, optional
        The port of the QuantumNetworkServer. The default is 7070.

    Examples
    --------

    We create a QuantumNetworkServer listening on the localhost IP-address and connect
    the client.

    >>> from qrisp.quantum_network import QuantumNetworkServer, QuantumNetworkClient
    >>> local_server = QuantumNetworkServer("127.0.0.1", background = True)
    >>> local_server.start()
    >>> client = QuantumNetworkClient(name = "alice", socket_ip = "127.0.0.1")


    """

    def __init__(self, name, socket_ip, port=None):
        from thrift.protocol import TBinaryProtocol
        from thrift.transport import TSocket, TTransport

        if port is None:
            port = 7070
        # Create the transport for the User Interface to the server
        self.transport = TSocket.TSocket(socket_ip, port)
        # Buffering is critical. Raw sockets are very slow
        self.transport = TTransport.TBufferedTransport(self.transport)
        # Wrap in a protocol
        protocol = TBinaryProtocol.TBinaryProtocol(self.transport)
        self.socket_ip = socket_ip
        # Create a client to use the protocol encoder

        super().__init__(protocol)

        # Connect!
        self.transport.open()

        self.name = name
        self.register_client(name)

    # Destructor closes the transport
    def __del__(self):
        self.transport.close()

    def run(self, qc, shots=None):
        r"""
        Runs a QuantumCircuit on the client's backend. Note that QuantumNetwork
        simulations do not support multiple shots, as the quantum state of the network
        is stored and updated everytime a client sends a query. Multiple shots could
        yield differing measurement outcomes, which implies an ambiguous quantum state
        of the network. Nevertheless, the results are returned in the form of a
        dictionary in order to comply with the quantum circuit execution backend
        interface.

        Note that it is possible to submit QuantumCircuits which contain qubits, that
        have not been requested previously. In this case, the qubits names are
        internally extended by the string "@client_name"  (if they aren't extended in
        this way already). This is to allow multiple clients to submit circuits with
        matching qubit names.

        Parameters
        ----------
        qc : QuantumCircuit
            The QuantumCircuit to run.

        Returns
        -------
        res : dict
            A dictionary containing a single key/value pair where the key represents the
            measurement outcome.


        Examples
        --------

        We create a local QuantumNetworkServer, connect a client and run a
        QuantumCircuit.

        >>> from qrisp.quantum_network import QuantumNetworkServer, QuantumNetworkClient
        >>> local_server = QuantumNetworkServer("127.0.0.1", background = True)
        >>> local_server.start()
        >>> client = QuantumNetworkClient(name = "alice", socket_ip = "127.0.0.1")
        >>> from qrisp import QuantumCircuit
        >>> qc = QuantumCircuit(2)
        >>> qc.h(0)
        >>> qc.cx(0,1)
        >>> qc.measure([0,1])
        >>> client.run(qc)
        {'11': 1}

        We perform another shot

        >>> client.run(qc)
        {'01': 1}

        After applying the first run command, the quantum state is

        .. math::
            \ket{\psi} = \ket{1}\ket{1}

        Applying the Hadamard yields

        .. math::
            \text{H}_0 \ket{\psi} = \frac{1}{\sqrt{2}}(\ket{0} - \ket{1})\ket{1}

        The CX yields

        .. math::
            \text{CX}_{01} \text{H}_0 \ket{\psi} =
            \frac{1}{\sqrt{2}}(\ket{0}\ket{1} - \ket{1}\ket{0})

        Finally, the measurement collapsed the state into the first summand.

        """
        converted_circuit = convert_circuit(qc, "thrift")

        res = super().run(converted_circuit, self.name)

        return res

        qc = super().get_clear_qc(self.name)

        return convert_circuit(qc, "qrisp")

    def send_qubits(self, recipient, qubits, annotation):
        """
        Sends the specified qubits to another participant of the network.

        Parameters
        ----------
        recipient : str
            The recipients name.
        qubits : list[Qubit]
            The list of qubits to send.
        annotation : str
            A string containing an arbitrary message that is available for the
            reciepient.


        Examples
        --------

        We create a QuantumNetworkServer, connect two clients and distribute a bell pair
        another client.

        >>> from qrisp.quantum_network import QuantumNetworkServer, QuantumNetworkClient
        >>> local_server = QuantumNetworkServer("127.0.0.1", background = True)
        >>> local_server.start()
        >>> alice_client = QuantumNetworkClient(name = "alice", socket_ip = "127.0.0.1")
        >>> bob_client = QuantumNetworkClient(name = "bob", socket_ip = "127.0.0.1")

        Now we create the bell pair:

        >>> from qrisp import QuantumCircuit
        >>> qc = QuantumCircuit(2)
        >>> qc.h(0)
        >>> qc.cx(0,1)
        >>> alice_client.run(qc)
        {'': 1}

        Send one of the qubits to Bob:

        >>> alice_client.send_qubits("bob", [qc.qubits[0]], annotation = "Happy birthday!")

        Now Bob can measure

        >>> messages = bob_client.inbox()
        >>> received_qubits, annotation = messages[0]
        >>> qc = bob_client.get_clear_qc()
        >>> qc.measure(received_qubits)
        >>> bob_client.run(qc)
        {'1': 1}

        After Bob's measurement, we expect Alice's measurement to yield the same due to
        `spukhafter Fernwirkung
        <https://en.wikipedia.org/wiki/Quantum_entanglement#History>`_:

        >>> qc = alice_client.get_clear_qc()
        >>> qc.measure(qc.qubits)
        >>> alice_client.run(qc)
        {'1': 1}



        """

        msg = Message([PortableQubit(qb.identifier) for qb in qubits], annotation)

        super().send_qubits(self.name, recipient, msg)

    def request_qubits(self, amount):
        """
        Creates the specified amount of qubits for the client's backend.

        Parameters
        ----------
        amount : int
            The amount of qubits to create.

        Returns
        -------
        list[Qubit]
            The qubits created.

        Examples
        --------

        >>> from qrisp.quantum_network import QuantumNetworkServer, QuantumNetworkClient
        >>> local_server = QuantumNetworkServer("127.0.0.1", background = True)
        >>> local_server.start()
        >>> client = QuantumNetworkClient(name = "alice", socket_ip = "127.0.0.1")
        >>> qb_list = client.request_qubits(4)
        >>> print(qb_list)
        [Qubit(qb_0@alice), Qubit(qb_1@alice), Qubit(qb_2@alice), Qubit(qb_3@alice)]

        """

        qb_list = super().request_qubits(amount, self.name)

        return [Qubit(qb.identifier) for qb in qb_list]

    def get_clear_qc(self):
        """
        Returns a QuantumCircuit containing all the qubits that belong to the client at
        the moment.

        Note that the qubit names of the run QuantumCircuit are internally extended by
        the string "@client_name"  (if they aren't extended in this way already). This
        is to allow multiple clients to submit circuits with matching qubit names.

        Note that

        Returns
        -------
        QuantumCircuit
            An empty QuantumCircuit containing the qubits, this client may operate on.

        Examples
        --------

        We set up a quantum network, run some QuantumCircuits and retrieve the clear
        QuantumCircuit:

        >>> from qrisp.quantum_network import QuantumNetworkServer, QuantumNetworkClient
        >>> local_server = QuantumNetworkServer("127.0.0.1", background = True)
        >>> local_server.start()
        >>> client = QuantumNetworkClient(name = "alice", socket_ip = "127.0.0.1")
        >>> from qrisp import QuantumCircuit
        >>> qc_0 = QuantumCircuit(1)
        >>> qc_0.x(0)
        >>> print(qc_0)
              ┌───┐
        qb_8: ┤ X ├
              └───┘
        >>> client.run(qc_0)
        >>> qc_1 = QuantumCircuit(1)
        >>> qc_1.h(0)
        >>> print(qc_1)
                ┌───┐
         qb_26: ┤ H ├
                └───┘
        >>> client.run(qc_1)
        >>> print(client.get_clear_qc().qubits)
        [Qubit(qb_58@alice), Qubit(qb_77@alice)]

        """
        qc = super().get_clear_qc(self.name)

        return convert_circuit(qc, "qrisp")

    def get_overall_qc(self):
        """
        Retrieves the QuantumCircuit of all operations that happened in the network so
        far.

        Returns
        -------
        QuantumCircuit
            The QuantumCircuit containing all operations that happened in this
            QuantumNetwork.

        Examples
        --------

        We assume that the commands of the example of the
        :meth:`send_qubits <qrisp.quantum_network.QuantumNetworkClient.send_qubits>`
        method have been executed.

        >>> print(alice_client.get_overall_qc())
                      ░ ┌───┐      ░ ┌─┐ ░
        qb_34@alice: ─░─┤ H ├──■───░─┤M├─░────
                      ░ └───┘┌─┴─┐ ░ └╥┘ ░ ┌─┐
        qb_35@alice: ─░──────┤ X ├─░──╫──░─┤M├
                      ░      └───┘ ░  ║  ░ └╥┘
           cb_bob_0: ═════════════════╩═════╬═
                                            ║
         cb_alice_0: ═══════════════════════╩═


        """

        qc = super().get_overall_qc()

        return convert_circuit(qc, "qrisp")

    def inbox(self):
        """
        Returns a list of tuples containing the received qubits and their annotations.

        Returns
        -------
        res_list : list[tuple[list[Qubit], str]]
            A list of tuples containing the sent qubits and the annotations.


        Examples
        --------

        We send a qubit from one client to another and inspect the inbox

        >>> from qrisp.quantum_network import QuantumNetworkServer, QuantumNetworkClient
        >>> local_server = QuantumNetworkServer("127.0.0.1", background = True)
        >>> local_server.start()
        >>> alice_client = QuantumNetworkClient(name = "alice", socket_ip = "127.0.0.1")
        >>> bob_client = QuantumNetworkClient(name = "bob", socket_ip = "127.0.0.1")

        Prepare the qubit which will be sent to bob:

        >>> from qrisp import QuantumCircuit
        >>> qc = QuantumCircuit(1)
        >>> qc.h(0)
        >>> alice_client.run(qc)
        {'': 1}

        Send the qubit to Bob:

        >>> alice_client.send_qubits("bob", [qc.qubits[0]], annotation = "Merry christmas!")

        Now Bob can check his inbox

        >>> messages = bob_client.inbox()
        >>> received_qubits, annotation = messages[0]
        >>> print(received_qubits)
        [Qubit(qb_20@alice)]
        >>> print(annotation)
        Merry christmas!

        """

        inbox_list = super().inbox(self.name)

        res_list = []

        for i in range(len(inbox_list)):
            res_list.append(
                (
                    [Qubit(qb.identifier) for qb in inbox_list[i].qubits],
                    inbox_list[i].annotation,
                )
            )
        return res_list
